from enum import Enum


class SoundFile(Enum):
    """
    Paths to standard EV3 sounds.

    Note:
        Expressions: BOING, BOO, CHEERING, CRUNCHING, CRYING, FANFARE, KUNG_FU, LAUGHING_1, LAUGHING_2, MAGIC_WAND, OUCH, SHOUTING, SMACK, SNEEZING, SNORING, UH_OH

        Information: ACTIVATE, ANALYZE, BACKWARDS, COLOR, DETECTED, DOWN, ERROR, ERROR_ALARM, FLASHING, FORWARD, LEFT, OBJECT, RIGHT, SEARCHING, START, STOP, TOUCH, TURN, UP

        Communication: BRAVO, EV3, FANTASTIC, GAME_OVER, GO, GOOD_JOB, GOOD, GOODBYE, HELLO, HI, LEGO, MINDSTORMS,MORNING, NO, OKAY, OKEY_DOKEY, SORRY, THANK_YOU, YES

        Movements: SPEED_DOWN, SPEED_IDLE, SPEED_UP

        Color: BLACK, BLUE, BROWN, GREEN, RED, WHITE, YELLOW

        Mechanical: AIR_RELEASE, AIRBRAKE, BACKING_ALERT, HORN_1, HORN_2, LASER, MOTOR_IDLE, MOTOR_START, MOTOR_STOP, RATCHET, SONAR, TICK_TACK

        Animal sounds: CAT_PURR, DOG_BARK_1, DOG_BARK_2, DOG_GROWL, DOG_SNIFF, DOG_WHINE, ELEPHANT_CALL, INSECT_BUZZ_1,INSECT_BUZZ_2, INSECT_CHIRP, SNAKE_HISS, SNAKE_RATTLE, T_REX_ROAR

        Numbers: ZERO, ONE, TWO, THREE, FOUR, FIVE, SIX, SEVEN, EIGHT, NINE, TEN

        System sounds: CLICK, CONFIRM, GENERAL_ALERT, OVERPOWER, READY
    """
    BOING = "BOING"
    BOO = "BOO"
    CHEERING = "CHEERING"
    CRUNCHING = "CRUNCHING"
    CRYING = "CRYING"
    FANFARE = "FANFARE"
    KUNG_FU = "KUNG_FU"
    LAUGHING_1 = "LAUGHING_1"
    LAUGHING_2 = "LAUGHING_2"
    MAGIC_WAND = "MAGIC_WAND"
    OUCH = "OUCH"
    SHOUTING = "SHOUTING"
    SMACK = "SMACK"
    SNEEZING = "SNEEZING"
    SNORING = "SNORING"
    UH_OH = "UH_OH"

    ACTIVATE = "ACTIVATE"
    ANALYZE = "ANALYZE"
    BACKWARDS = "BACKWARDS"
    COLOR = "COLOR"
    DETECTED = "DETECTED"
    DOWN = "DOWN"
    ERROR = "ERROR"
    ERROR_ALARM = "ERROR_ALARM"
    FLASHING = "FLASHING"
    FORWARD = "FORWARD"
    LEFT = "LEFT"
    OBJECT = "OBJECT"
    RIGHT = "RIGHT"
    SEARCHING = "SEARCHING"
    START = "START"
    STOP = "STOP"
    TOUCH = "TOUCH"
    TURN = "TURN"
    UP = "UP"

    BRAVO = "BRAVO"
    EV3 = "EV3"
    FANTASTIC = "FANTASTIC"
    GAME_OVER = "GAME_OVER"
    GO = "GO"
    GOOD_JOB = "GOOD_JOB"
    GOOD = "GOOD"
    GOODBYE = "GOODBYE"
    HELLO = "HELLO"
    HI = "HI"
    LEGO = "LEGO"
    MINDSTORMS = "MINDSTORMS"
    MORNING = "MORNING"
    NO = "NO"
    OKAY = "OKAY"
    OKEY_DOKEY = "OKEY_DOKEY"
    SORRY = "SORRY"
    THANK_YOU = "THANK_YOU"
    YES = "YES"

    SPEED_DOWN = "SPEED_DOWN"
    SPEED_IDLE = "SPEED_IDLE"
    SPEED_UP = "SPEED_UP"

    BLACK = "BLACK"
    BLUE = "BLUE"
    BROWN = "BROWN"
    GREEN = "GREEN"
    RED = "RED"
    WHITE = "WHITE"
    YELLOW = "YELLOW"

    AIR_RELEASE = "AIR_RELEASE"
    AIRBRAKE = "AIRBRAKE"
    BACKING_ALERT = "BACKING_ALERT"
    HORN_1 = "HORN_1"
    HORN_2 = "HORN_2"
    LASER = "LASER"
    MOTOR_IDLE = "MOTOR_IDLE"
    MOTOR_START = "MOTOR_START"
    MOTOR_STOP = "MOTOR_STOP"
    RATCHET = "RATCHET"
    SONAR = "SONAR"
    TICK_TACK = "TICK_TACK"

    CAT_PURR = "CAT_PURR"
    DOG_BARK_1 = "DOG_BARK_1"
    DOG_BARK_2 = "DOG_BARK_2"
    DOG_GROWL = "DOG_GROWL"
    DOG_SNIFF = "DOG_SNIFF"
    DOG_WHINE = "DOG_WHINE"
    ELEPHANT_CALL = "ELEPHANT_CALL"
    INSECT_BUZZ_1 = "INSECT_BUZZ_1"
    INSECT_BUZZ_2 = "INSECT_BUZZ_2"
    INSECT_CHIRP = "INSECT_CHIRP"
    SNAKE_HISS = "SNAKE_HISS"
    SNAKE_RATTLE = "SNAKE_RATTLE"
    T_REX_ROAR = "T_REX_ROAR"

    ZERO = "ZERO"
    ONE = "ONE"
    TWO = "TWO"
    THREE = "THREE"
    FOUR = "FOUR"
    FIVE = "FIVE"
    SIX = "SIX"
    SEVEN = "SEVEN"
    EIGHT = "EIGHT"
    NINE = "NINE"
    TEN = "TEN"

    CLICK = "CLICK"
    CONFIRM = "CONFIRM"
    GENERAL_ALERT = "GENERAL_ALERT"
    OVERPOWER = "OVERPOWER"
    READY = "READY"



import json
import os.path

import polib
import pyperclip

locale_dir = os.path.dirname(__file__)


def register_locale(name: str) -> None:
    print(f"Processing {name}")
    po_path = os.path.join(locale_dir, name, "LC_MESSAGES", "turcar.po")
    po = polib.pofile(po_path)

    registered_path = os.path.join(locale_dir, name, "LC_MESSAGES", "registered.json")
    if not os.path.exists(registered_path):
        with open(registered_path, "w", encoding="utf-8") as fp:
            fp.write("{}")

    with open(registered_path, encoding="utf-8") as fp:
        registered = json.load(fp)

    new_registered = {}

    review_messages = []

    for entry in po:
        if entry.msgstr and (
            entry.msgid not in registered or registered[entry.msgid] != entry.msgstr
        ):
            msg = entry.msgstr.strip().replace("\n", " ")
            if not msg.endswith("."):
                msg = msg + "."

            review_messages.append(msg)

        if entry.msgstr:
            new_registered[entry.msgid] = entry.msgstr

    if review_messages:
        print("\n".join(review_messages))
        pyperclip.copy("\n".join(review_messages))

        input(f"... Press ENTER to confirm {name}! ...")

    with open(registered_path, "w", encoding="utf-8") as fp:
        json.dump(new_registered, fp, sort_keys=True, indent=4, ensure_ascii=False)

    print("--------------------------------------")


for name in os.listdir(locale_dir):
    path = os.path.join(locale_dir, name)
    if os.path.isdir(path):
        register_locale(name)



"""
asynchronous I/O scheduler for writing concurrent code.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/uasyncio.rst.

|see_cpython_module|
`asyncio <https://docs.python.org/3.8/library/asyncio.html>`_

Example::

    import uasyncio

    async def blink(led, period_ms):
        while True:
            led.on()
            await uasyncio.sleep_ms(5)
            led.off()
            await uasyncio.sleep_ms(period_ms)

    async def main(led1, led2):
        uasyncio.create_task(blink(led1, 700))
        uasyncio.create_task(blink(led2, 400))
        await uasyncio.sleep_ms(10_000)

    # Running on a pyboard
    from pyb import LED
    uasyncio.run(main(LED(1), LED(2)))

    # Running on a generic board
    from machine import Pin
    uasyncio.run(main(Pin(1), Pin(2)))
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from abc import ABC
from typing import Awaitable, TypeVar, Callable
from typing import Coroutine, Any, Dict, Iterable, Generic, Final

from uio import AnyReadableBuf

_T: Final = TypeVar("_T")
# `Coroutine` `_T` is covariant and `Awaitable` `_T` is invariant.
_C: Final = Coroutine[Any, None, _T] | Awaitable[_T]

class CancelledError(BaseException):
    """
   The operation has been cancelled.

   This exception can be caught to perform custom operations when asyncio Tasks are cancelled. 
   In almost all situations the exception must be re-raised.
   """

class IOQueue:
    """
   Queue and poller for stream IO.
   """

class SingletonGenerator:
    """
   `Yield` once, then raise StopIteration.
   """

class TimeoutError(Exception):
    """
   The operation has exceeded the given deadline.
   """

def run_until_complete(main_task: Awaitable[_T] | None = None, /):
    """
   Keep scheduling tasks until there are none left to schedule.
   """

def create_task(coro: _C, /) -> Task[_T]:
    """
    Create a new task from the given coroutine and schedule it to run.
    
    Returns the corresponding `Task` object.
   """

def current_task() -> Task[Any] | None:
    """
    Return the `Task` object associated with the currently running task.
   """

def run(coro: _C, /) -> _T:
    """
    Create a new task from the given coroutine and run it until it completes.
    
    Returns the value returned by *coro*.
   """

def sleep(t: float, /) -> Awaitable[None]:
    """
    Sleep for *t* seconds (can be a float).
    
    This is a coroutine.
   """

def sleep_ms(t: int, /) -> Awaitable[None]:
    """
    Sleep for *t* milliseconds.
    
    This is a coroutine, and a MicroPython extension.
   """

def wait_for(awaitable: Awaitable[_T], timeout: float, /) -> Awaitable[_T]:
    """
    Wait for the *awaitable* to complete, but cancel it if it takes longer
    than *timeout* seconds.  If *awaitable* is not a task then a task will be
    created from it.
    
    If a timeout occurs, it cancels the task and raises ``asyncio.TimeoutError``:
    this should be trapped by the caller.  The task receives
    ``asyncio.CancelledError`` which may be ignored or trapped using ``try...except``
    or ``try...finally`` to run cleanup code.
    
    Returns the return value of *awaitable*.
    
    This is a coroutine.
   """

def wait_for_ms(awaitable: Awaitable[_T], timeout: int, /) -> Awaitable[_T]:
    """
    Similar to `wait_for` but *timeout* is an integer in milliseconds.
    
    This is a coroutine, and a MicroPython extension.
   """

def gather(
    *awaitable: Awaitable[Any], return_exceptions: bool = False
) -> Awaitable[list[Any]]:
    """
    Run all *awaitables* concurrently.  Any *awaitables* that are not tasks are
    promoted to tasks.
    
    Returns a list of return values of all *awaitables*.
    
    This is a coroutine.
   """

StreamReader = "Stream"

StreamWriter = "Stream"

def open_connection(
    host: str | None, port: str | int | None, /,
) -> Awaitable[tuple[StreamReader, StreamWriter]]:
    """
    Open a TCP connection to the given *host* and *port*.  The *host* address will be
    resolved using `socket.getaddrinfo`, which is currently a blocking call.
    
    Returns a pair of streams: a reader and a writer stream.
    Will raise a socket-specific ``OSError`` if the host could not be resolved or if
    the connection could not be made.
    
    This is a coroutine.
   """

def start_server(
    callback: Callable[[StreamReader, StreamWriter], None],
    host: str | None,
    port: str | int | None,
    backlog: int = 5,
    /,
) -> Awaitable[Server]:
    """
    Start a TCP server on the given *host* and *port*.  The *callback* will be
    called with incoming, accepted connections, and be passed 2 arguments: reader
    and writer streams for the connection.
    
    Returns a `Server` object.
    
    This is a coroutine.
   """

def get_event_loop() -> Loop:
    """
    Return the event loop used to schedule and run tasks.  See `Loop`.
   """

def new_event_loop() -> Loop:
    """
    Reset the event loop and return it.
    
    Note: since MicroPython only has a single event loop this function just
    resets the loop's state, it does not create a new one.
   """

class Task(Awaitable[_T], Iterable[_T], Generic[_T], ABC):
    """
   class Task
   ----------
   """

    def __init__(self):
        """
       This object wraps a coroutine into a running task.  Tasks can be waited on
       using ``await task``, which will wait for the task to complete and return
       the return value of the task.
       
       Tasks should not be created directly, rather use `create_task` to create them.
      """
    def cancel(self) -> None:
        """
       Cancel the task by injecting ``asyncio.CancelledError`` into it.  The task may
       ignore this exception.  Cleanup code may be run by trapping it, or via
       ``try ... finally``.
      """

class Event:
    """
   class Event
   -----------
   """

    def __init__(self):
        """
       Create a new event which can be used to synchronise tasks.  Events start
       in the cleared state.
      """
    def is_set(self) -> bool:
        """
       Returns ``True`` if the event is set, ``False`` otherwise.
      """
    def set(self) -> None:
        """
       Set the event.  Any tasks waiting on the event will be scheduled to run.
       
       Note: This must be called from within a task. It is not safe to call this
       from an IRQ, scheduler callback, or other thread. See `ThreadSafeFlag`.
      """
    def clear(self) -> None:
        """
       Clear the event.
      """
    def wait(self) -> Awaitable[Any]:
        """
       Wait for the event to be set.  If the event is already set then it returns
       immediately.
       
       This is a coroutine.
      """

class ThreadSafeFlag:
    """
   class ThreadSafeFlag
   --------------------
   """

    def __init__(self):
        """
       Create a new flag which can be used to synchronise a task with code running
       outside the asyncio loop, such as other threads, IRQs, or scheduler
       callbacks.  Flags start in the cleared state.
      """
    def set(self) -> None:
        """
       Set the flag.  If there is a task waiting on the event, it will be scheduled
       to run.
      """
    def wait(self) -> Awaitable[None]:
        """
       Wait for the flag to be set.  If the flag is already set then it returns
       immediately.
       
       A flag may only be waited on by a single task at a time.
       
       This is a coroutine.
      """

class Lock(Awaitable[None], ABC):
    """
   class Lock
   ----------
   """

    def __init__(self):
        """
       Create a new lock which can be used to coordinate tasks.  Locks start in
       the unlocked state.
       
       In addition to the methods below, locks can be used in an ``async with`` statement.
      """
    def locked(self) -> bool:
        """
       Returns ``True`` if the lock is locked, otherwise ``False``.
      """
    def acquire(self) -> Awaitable[None]:
        """
       Wait for the lock to be in the unlocked state and then lock it in an atomic
       way.  Only one task can acquire the lock at any one time.
       
       This is a coroutine.
      """
    def release(self) -> None:
        """
       Release the lock.  If any tasks are waiting on the lock then the next one in the
       queue is scheduled to run and the lock remains locked.  Otherwise, no tasks are
       waiting an the lock becomes unlocked.
      """

class Stream:
    """

   """

    def __init__(self):
        """
       This represents a TCP stream connection.  To minimise code this class implements
       both a reader and a writer, and both ``StreamReader`` and ``StreamWriter`` alias to
       this class.
      """
    def get_extra_info(self, v: str, /) -> str:
        """
       Get extra information about the stream, given by *v*.  The valid values for *v* are:
       ``peername``.
      """
    def close(self) -> None:
        """
       Close the stream.
      """
    def wait_close(self) -> Awaitable[None]:
        """
       Wait for the stream to close.
       
       This is a coroutine.
      """
    def read(self, n: int, /) -> Awaitable[bytes]:
        """
       Read up to *n* bytes and return them.
       
       This is a coroutine.
      """
    def readline(self) -> Awaitable[bytes]:
        """
       Read a line and return it.
       
       This is a coroutine.
      """
    def write(self, buf: AnyReadableBuf, /) -> None:
        """
       Accumulated *buf* to the output buffer.  The data is only flushed when
       `Stream.drain` is called.  It is recommended to call `Stream.drain` immediately
       after calling this function.
      """
    def drain(self) -> Awaitable[None]:
        """
       Drain (write) all buffered output data out to the stream.
       
       This is a coroutine.
      """

class Server:
    """

   """

    def __init__(self):
        """
       This represents the server class returned from `start_server`.  It can be used
       in an ``async with`` statement to close the server upon exit.
      """
    def close(self) -> None:
        """
       Close the server.
      """
    def wait_close(self) -> Awaitable[None]:
        """
       Wait for the server to close.
       
       This is a coroutine.
      """

class Loop:
    """

   """

    def __init__(self):
        """
       This represents the object which schedules and runs tasks.  It cannot be
       created, use `get_event_loop` instead.
      """
    def create_task(self, coro: _C, /) -> Task[_T]:
        """
       Create a task from the given *coro* and return the new `Task` object.
      """
    def run_forever(self) -> None:
        """
       Run the event loop until `stop()` is called.
      """
    def run_until_complete(self, awaitable: Awaitable[_T], /) -> None:
        """
       Run the given *awaitable* until it completes.  If *awaitable* is not a task
       then it will be promoted to one.
      """
    def stop(self) -> None:
        """
       Stop the event loop.
      """
    def close(self) -> None:
        """
       Close the event loop.
      """
    def set_exception_handler(
        self, handler: Callable[[Loop, Dict[str, Any]], None] | None, /
    ) -> None:
        """
       Set the exception handler to call when a Task raises an exception that is not
       caught.  The *handler* should accept two arguments: ``(loop, context)``.
      """
    def get_exception_handler(self) -> Callable[[Loop, Dict[str, Any]], None] | None:
        """
       Get the current exception handler.  Returns the handler, or ``None`` if no
       custom handler is set.
      """
    def default_exception_handler(self, context: Dict[str, Any], /) -> None:
        """
       The default exception handler that is called.
      """
    def call_exception_handler(self, context: Dict[str, Any], /) -> None:
        """
       Call the current exception handler.  The argument *context* is passed through and
       is a dictionary containing keys: ``'message'``, ``'exception'``, ``'future'``.
      """



"""
basic "operating system" services.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/os.rst.
==============================================

.. module:: os
   :synopsis: basic "operating system" services

|see_cpython_module| :mod:`python:os`.

The ``os`` module contains functions for filesystem access and mounting,
terminal redirection and duplication, and the ``uname`` and ``urandom``
functions.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from abc import abstractmethod
from typing import Final, TypeVar, runtime_checkable, Protocol, overload, Literal

from uio import IOBase

_StrOrBytesT: Final = TypeVar("_StrOrBytesT", str, bytes)

class _PathLike(Protocol[_StrOrBytesT]):
    @abstractmethod
    def __fspath__(self) -> _StrOrBytesT:
        """Return the file system path representation of the object, preferably as a `str`."""

_AnyPath: Final = str | bytes | _PathLike[str] | _PathLike[bytes]
_FdOrAnyPath: Final = int | _AnyPath

def uname() -> tuple[str, str, str, str, str]:
    """
   Return a tuple (possibly a named tuple) containing information about the
   underlying machine and/or its operating system.  The tuple has five fields
   in the following order, each of them being a string:
   
        * ``sysname`` -- the name of the underlying system
        * ``nodename`` -- the network name (can be the same as ``sysname``)
        * ``release`` -- the version of the underlying system
        * ``version`` -- the MicroPython version and build date
        * ``machine`` -- an identifier for the underlying hardware (eg board, CPU)
   """

def urandom(n: int, /) -> bytes:
    """
   Return a bytes object with *n* random bytes. Whenever possible, it is
   generated by the hardware random number generator.
   """

def chdir(path: _FdOrAnyPath, /) -> None:
    """
   Change current directory.
   """

def getcwd() -> str:
    """
   Get the current directory.
   """

@overload
def ilistdir() -> list[tuple[str, int, int] | tuple[str, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def ilistdir(dir: int, /) -> list[tuple[str, int, int] | tuple[str, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def ilistdir(dir: str, /) -> list[tuple[str, int, int] | tuple[str, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def ilistdir(
    dir: bytes, /
) -> list[tuple[bytes, int, int] | tuple[bytes, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def ilistdir(
    dir: _PathLike[str], /
) -> list[tuple[str, int, int] | tuple[str, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def ilistdir(
    dir: _PathLike[bytes], /
) -> list[tuple[bytes, int, int] | tuple[bytes, int, int, int]]:
    """
   This function returns an iterator which then yields tuples corresponding to
   the entries in the directory that it is listing.  With no argument it lists the
   current directory, otherwise it lists the directory given by *dir*.
   
   The tuples have the form *(name, type, inode[, size])*:
   
    - *name* is a string (or bytes if *dir* is a bytes object) and is the name of
      the entry;
    - *type* is an integer that specifies the type of the entry, with 0x4000 for
      directories and 0x8000 for regular files;
    - *inode* is an integer corresponding to the inode of the file, and may be 0
      for filesystems that don't have such a notion.
    - Some platforms may return a 4-tuple that includes the entry's *size*.  For
      file entries, *size* is an integer representing the size of the file
      or -1 if unknown.  Its meaning is currently undefined for directory
      entries.
   """

@overload
def listdir() -> list[str]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

@overload
def listdir(dir: int, /) -> list[str]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

@overload
def listdir(dir: str, /) -> list[str]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

@overload
def listdir(dir: bytes, /) -> list[bytes]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

@overload
def listdir(dir: _PathLike[str], /) -> list[str]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

@overload
def listdir(dir: _PathLike[bytes], /) -> list[bytes]:
    """
   With no argument, list the current directory.  Otherwise list the given directory.
   """

def mkdir(path: _AnyPath, /) -> None:
    """
   Create a new directory.
   """

def remove(path: _AnyPath, /) -> None:
    """
   Remove a file.
   """

def rmdir(path: _AnyPath, /) -> None:
    """
   Remove a directory.
   """

def rename(old_path: _AnyPath, new_path: _AnyPath, /) -> None:
    """
   Rename a file.
   """

def stat(
    path: _FdOrAnyPath, /
) -> tuple[int, int, int, int, int, int, int, int, int, int]:
    """
   Get the status of a file or directory.
   """

def statvfs(
    path: _FdOrAnyPath, /
) -> tuple[int, int, int, int, int, int, int, int, int, int]:
    """
   Get the status of a fileystem.
   
   Returns a tuple with the filesystem information in the following order:
   
        * ``f_bsize`` -- file system block size
        * ``f_frsize`` -- fragment size
        * ``f_blocks`` -- size of fs in f_frsize units
        * ``f_bfree`` -- number of free blocks
        * ``f_bavail`` -- number of free blocks for unprivileged users
        * ``f_files`` -- number of inodes
        * ``f_ffree`` -- number of free inodes
        * ``f_favail`` -- number of free inodes for unprivileged users
        * ``f_flag`` -- mount flags
        * ``f_namemax`` -- maximum filename length
   
   Parameters related to inodes: ``f_files``, ``f_ffree``, ``f_avail``
   and the ``f_flags`` parameter may return ``0`` as they can be unavailable
   in a port-specific implementation.
   """

def sync() -> None:
    """
   Sync all filesystems.
   """

def dupterm(stream_object: IOBase | None, index: int = 0, /) -> IOBase | None:
    """
   Duplicate or switch the MicroPython terminal (the REPL) on the given `stream`-like
   object. The *stream_object* argument must be a native stream object, or derive
   from ``io.IOBase`` and implement the ``readinto()`` and
   ``write()`` methods.  The stream should be in non-blocking mode and
   ``readinto()`` should return ``None`` if there is no data available for reading.
   
   After calling this function all terminal output is repeated on this stream,
   and any input that is available on the stream is passed on to the terminal input.
   
   The *index* parameter should be a non-negative integer and specifies which
   duplication slot is set.  A given port may implement more than one slot (slot 0
   will always be available) and in that case terminal input and output is
   duplicated on all the slots that are set.
   
   If ``None`` is passed as the *stream_object* then duplication is cancelled on
   the slot given by *index*.
   
   The function returns the previous stream-like object in the given slot.
   """

def mount(
    fsobj: "AbstractBlockDev", mount_point: str, /, *, readonly: bool = False
) -> IOBase | None:
    """
    Mount the filesystem object *fsobj* at the location in the VFS given by the
    *mount_point* string.  *fsobj* can be a a VFS object that has a ``mount()``
    method, or a block device.  If it's a block device then the filesystem type
    is automatically detected (an exception is raised if no filesystem was
    recognised).  *mount_point* may be ``'/'`` to mount *fsobj* at the root,
    or ``'/<name>'`` to mount it at a subdirectory under the root.
    
    If *readonly* is ``True`` then the filesystem is mounted read-only.
    
    During the mount process the method ``mount()`` is called on the filesystem
    object.
    
    Will raise ``OSError(EPERM)`` if *mount_point* is already mounted.
    
    
    Filesystem mounting
    -------------------
    
    Some ports provide a Virtual Filesystem (VFS) and the ability to mount multiple
    "real" filesystems within this VFS.  Filesystem objects can be mounted at either
    the root of the VFS, or at a subdirectory that lives in the root.  This allows
    dynamic and flexible configuration of the filesystem that is seen by Python
    programs.  Ports that have this functionality provide the :func:`mount` and
    :func:`umount` functions, and possibly various filesystem implementations
    represented by VFS classes.
   """

def umount(mount_point: str, /) -> None:
    """
    Unmount a filesystem. *mount_point* can be a string naming the mount location,
    or a previously-mounted filesystem object.  During the unmount process the
    method ``umount()`` is called on the filesystem object.
    
    Will raise ``OSError(EINVAL)`` if *mount_point* is not found.
   """

class VfsFat("AbstractBlockDev"):
    """

   """

    def __init__(self, block_dev: "AbstractBlockDev", /):
        """
       Create a filesystem object that uses the FAT filesystem format.  Storage of
       the FAT filesystem is provided by *block_dev*.
       Objects created by this constructor can be mounted using :func:`mount`.
      """
    @staticmethod
    def mkfs(block_dev: "AbstractBlockDev", /) -> None:
        """
           Build a FAT filesystem on *block_dev*.
      """

class VfsLfs1("AbstractBlockDev"):
    """

   """

    def __init__(
        self,
        block_dev: "AbstractBlockDev",
        readsize: int = 32,
        progsize: int = 32,
        lookahead: int = 32,
        /,
    ):
        """
       Create a filesystem object that uses the `littlefs v1 filesystem format`_.
       Storage of the littlefs filesystem is provided by *block_dev*, which must
       support the :ref:`extended interface <block-device-interface>`.
       Objects created by this constructor can be mounted using :func:`mount`.
       
       See :ref:`filesystem` for more information.
      """
    @staticmethod
    def mkfs(
        block_dev: "AbstractBlockDev",
        readsize: int = 32,
        progsize: int = 32,
        lookahead: int = 32,
        /,
    ) -> None:
        """
           Build a Lfs1 filesystem on *block_dev*.
           
           .. note:: There are reports of littlefs v1 failing in certain situations,
                 for details see `littlefs issue 347`_.
      """

class VfsLfs2("AbstractBlockDev"):
    """

   """

    def __init__(
        self,
        block_dev: "AbstractBlockDev",
        readsize: int = 32,
        progsize: int = 32,
        lookahead: int = 32,
        mtime: bool = True,
        /,
    ):
        """
       Create a filesystem object that uses the `littlefs v2 filesystem format`_.
       Storage of the littlefs filesystem is provided by *block_dev*, which must
       support the :ref:`extended interface <block-device-interface>`.
       Objects created by this constructor can be mounted using :func:`mount`.
       
       The *mtime* argument enables modification timestamps for files, stored using
       littlefs attributes.  This option can be disabled or enabled differently each
       mount time and timestamps will only be added or updated if *mtime* is enabled,
       otherwise the timestamps will remain untouched.  Littlefs v2 filesystems without
       timestamps will work without reformatting and timestamps will be added
       transparently to existing files once they are opened for writing.  When *mtime*
       is enabled `os.stat` on files without timestamps will return 0 for the timestamp.
       
       See :ref:`filesystem` for more information.
      """
    @staticmethod
    def mkfs(
        block_dev: "AbstractBlockDev",
        readsize: int = 32,
        progsize: int = 32,
        lookahead: int = 32,
        mtime: bool = True,
        /,
    ) -> None:
        """
           Build a Lfs2 filesystem on *block_dev*.
           
           .. note:: There are reports of littlefs v2 failing in certain situations,
                 for details see `littlefs issue 295`_.
      """

@runtime_checkable
class AbstractBlockDev(Protocol):
    """
   Block devices
   -------------
   
   A block device is an object which implements the block protocol. This enables a
   device to support MicroPython filesystems. The physical hardware is represented
   by a user defined class. The :class:`AbstractBlockDev` class is a template for
   the design of such a class: MicroPython does not actually provide that class,
   but an actual block device class must implement the methods described below.
   
   A concrete implementation of this class will usually allow access to the
   memory-like functionality of a piece of hardware (like flash memory). A block
   device can be formatted to any supported filesystem and mounted using ``os``
   methods.
   
   See :ref:`filesystem` for example implementations of block devices using the
   two variants of the block protocol described below.
   
   .. _block-device-interface:
   
   Simple and extended interface
   .............................
   
   There are two compatible signatures for the ``readblocks`` and ``writeblocks``
   methods (see below), in order to support a variety of use cases.  A given block
   device may implement one form or the other, or both at the same time. The second
   form (with the offset parameter) is referred to as the "extended interface".
   
   Some filesystems (such as littlefs) that require more control over write
   operations, for example writing to sub-block regions without erasing, may require
   that the block device supports the extended interface.
   """

    def __init__(self):
        """
       Construct a block device object.  The parameters to the constructor are
       dependent on the specific block device.
      """
    @overload
    def readblocks(self, block_num: int, buf: bytearray, /) -> None:
        """
           The first form reads aligned, multiples of blocks.
           Starting at the block given by the index *block_num*, read blocks from
           the device into *buf* (an array of bytes).
           The number of blocks to read is given by the length of *buf*,
           which will be a multiple of the block size.
           
           The second form allows reading at arbitrary locations within a block,
           and arbitrary lengths.
           Starting at block index *block_num*, and byte offset within that block
           of *offset*, read bytes from the device into *buf* (an array of bytes).
           The number of bytes to read is given by the length of *buf*.
      """
    @overload
    def readblocks(self, block_num: int, buf: bytearray, offset: int, /) -> None:
        """
           The first form reads aligned, multiples of blocks.
           Starting at the block given by the index *block_num*, read blocks from
           the device into *buf* (an array of bytes).
           The number of blocks to read is given by the length of *buf*,
           which will be a multiple of the block size.
           
           The second form allows reading at arbitrary locations within a block,
           and arbitrary lengths.
           Starting at block index *block_num*, and byte offset within that block
           of *offset*, read bytes from the device into *buf* (an array of bytes).
           The number of bytes to read is given by the length of *buf*.
      """
    @overload
    def writeblocks(self, block_num: int, buf: bytes | bytearray, /) -> None:
        """
           The first form writes aligned, multiples of blocks, and requires that the
           blocks that are written to be first erased (if necessary) by this method.
           Starting at the block given by the index *block_num*, write blocks from
           *buf* (an array of bytes) to the device.
           The number of blocks to write is given by the length of *buf*,
           which will be a multiple of the block size.
           
           The second form allows writing at arbitrary locations within a block,
           and arbitrary lengths.  Only the bytes being written should be changed,
           and the caller of this method must ensure that the relevant blocks are
           erased via a prior ``ioctl`` call.
           Starting at block index *block_num*, and byte offset within that block
           of *offset*, write bytes from *buf* (an array of bytes) to the device.
           The number of bytes to write is given by the length of *buf*.
           
           Note that implementations must never implicitly erase blocks if the offset
           argument is specified, even if it is zero.
      """
    @overload
    def writeblocks(
        self, block_num: int, buf: bytes | bytearray, offset: int, /
    ) -> None:
        """
           The first form writes aligned, multiples of blocks, and requires that the
           blocks that are written to be first erased (if necessary) by this method.
           Starting at the block given by the index *block_num*, write blocks from
           *buf* (an array of bytes) to the device.
           The number of blocks to write is given by the length of *buf*,
           which will be a multiple of the block size.
           
           The second form allows writing at arbitrary locations within a block,
           and arbitrary lengths.  Only the bytes being written should be changed,
           and the caller of this method must ensure that the relevant blocks are
           erased via a prior ``ioctl`` call.
           Starting at block index *block_num*, and byte offset within that block
           of *offset*, write bytes from *buf* (an array of bytes) to the device.
           The number of bytes to write is given by the length of *buf*.
           
           Note that implementations must never implicitly erase blocks if the offset
           argument is specified, even if it is zero.
      """
    @overload
    def ioctl(self, op: int, arg: int, /) -> int | None:
        """
           Control the block device and query its parameters.  The operation to
           perform is given by *op* which is one of the following integers:
           
             - 1 -- initialise the device (*arg* is unused)
             - 2 -- shutdown the device (*arg* is unused)
             - 3 -- sync the device (*arg* is unused)
             - 4 -- get a count of the number of blocks, should return an integer
               (*arg* is unused)
             - 5 -- get the number of bytes in a block, should return an integer,
               or ``None`` in which case the default value of 512 is used
               (*arg* is unused)
             - 6 -- erase a block, *arg* is the block number to erase
           
           As a minimum ``ioctl(4, ...)`` must be intercepted; for littlefs
           ``ioctl(6, ...)`` must also be intercepted. The need for others is
           hardware dependent.
           
           Prior to any call to ``writeblocks(block, ...)`` littlefs issues
           ``ioctl(6, block)``. This enables a device driver to erase the block
           prior to a write if the hardware requires it. Alternatively a driver
           might intercept ``ioctl(6, block)`` and return 0 (success). In this case
           the driver assumes responsibility for detecting the need for erasure.
           
           Unless otherwise stated ``ioctl(op, arg)`` can return ``None``.
           Consequently an implementation can ignore unused values of ``op``. Where
           ``op`` is intercepted, the return value for operations 4 and 5 are as
           detailed above. Other operations should return 0 on success and non-zero
           for failure, with the value returned being an ``OSError`` errno code.
      """
    @overload
    def ioctl(self, op: Literal[4, 5], arg: int, /) -> int:
        """
           Control the block device and query its parameters.  The operation to
           perform is given by *op* which is one of the following integers:
           
             - 1 -- initialise the device (*arg* is unused)
             - 2 -- shutdown the device (*arg* is unused)
             - 3 -- sync the device (*arg* is unused)
             - 4 -- get a count of the number of blocks, should return an integer
               (*arg* is unused)
             - 5 -- get the number of bytes in a block, should return an integer,
               or ``None`` in which case the default value of 512 is used
               (*arg* is unused)
             - 6 -- erase a block, *arg* is the block number to erase
           
           As a minimum ``ioctl(4, ...)`` must be intercepted; for littlefs
           ``ioctl(6, ...)`` must also be intercepted. The need for others is
           hardware dependent.
           
           Prior to any call to ``writeblocks(block, ...)`` littlefs issues
           ``ioctl(6, block)``. This enables a device driver to erase the block
           prior to a write if the hardware requires it. Alternatively a driver
           might intercept ``ioctl(6, block)`` and return 0 (success). In this case
           the driver assumes responsibility for detecting the need for erasure.
           
           Unless otherwise stated ``ioctl(op, arg)`` can return ``None``.
           Consequently an implementation can ignore unused values of ``op``. Where
           ``op`` is intercepted, the return value for operations 4 and 5 are as
           detailed above. Other operations should return 0 on success and non-zero
           for failure, with the value returned being an ``OSError`` errno code.
      """
    @overload
    def ioctl(self, op: Literal[1, 2, 3, 6], arg: int, /) -> int | None:
        """
           Control the block device and query its parameters.  The operation to
           perform is given by *op* which is one of the following integers:
           
             - 1 -- initialise the device (*arg* is unused)
             - 2 -- shutdown the device (*arg* is unused)
             - 3 -- sync the device (*arg* is unused)
             - 4 -- get a count of the number of blocks, should return an integer
               (*arg* is unused)
             - 5 -- get the number of bytes in a block, should return an integer,
               or ``None`` in which case the default value of 512 is used
               (*arg* is unused)
             - 6 -- erase a block, *arg* is the block number to erase
           
           As a minimum ``ioctl(4, ...)`` must be intercepted; for littlefs
           ``ioctl(6, ...)`` must also be intercepted. The need for others is
           hardware dependent.
           
           Prior to any call to ``writeblocks(block, ...)`` littlefs issues
           ``ioctl(6, block)``. This enables a device driver to erase the block
           prior to a write if the hardware requires it. Alternatively a driver
           might intercept ``ioctl(6, block)`` and return 0 (success). In this case
           the driver assumes responsibility for detecting the need for erasure.
           
           Unless otherwise stated ``ioctl(op, arg)`` can return ``None``.
           Consequently an implementation can ignore unused values of ``op``. Where
           ``op`` is intercepted, the return value for operations 4 and 5 are as
           detailed above. Other operations should return 0 on success and non-zero
           for failure, with the value returned being an ``OSError`` errno code.
      """



"""
cpython_backend and cpython_frontend are separated, because the cpython_ssh needs to copy the former
to the remote host so that it wouldn't depend on frontend part.
"""
from .cp_back import MainCPythonBackend, get_backend



"""
Module: '_api.lightmatrix' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any

FACES = [] # type: list

class LightMatrix():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    def get_orientation(self, *args, **kwargs) -> Any:
        """"""

    def off(self, *args, **kwargs) -> Any:
        """"""

    def play_animation(self, *args, **kwargs) -> Any:
        """"""

    def rotate(self, *args, **kwargs) -> Any:
        """"""

    def set_orientation(self, *args, **kwargs) -> Any:
        """"""

    def set_pixel(self, *args, **kwargs) -> Any:
        """"""

    def show(self, *args, **kwargs) -> Any:
        """"""

    def show_image(self, *args, **kwargs) -> Any:
        """"""

    def start_animation(self, *args, **kwargs) -> Any:
        """"""

    def write(self, *args, **kwargs) -> Any:
        """"""

ORIENTATIONS = [] # type: list
def get_current_orientation(*args, **kwargs) -> Any:
    """"""

def rotate_orientation(*args, **kwargs) -> Any:
    """"""

def set_current_orientation(*args, **kwargs) -> Any:
    """"""




"""
simple BTree database.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/btree.rst.
=====================================

.. module:: btree
   :synopsis: simple BTree database

The ``btree`` module implements a simple key-value database using external
storage (disk files, or in general case, a random-access `stream`). Keys are
stored sorted in the database, and besides efficient retrieval by a key
value, a database also supports efficient ordered range scans (retrieval
of values with the keys in a given range). On the application interface
side, BTree database work as close a possible to a way standard `dict`
type works, one notable difference is that both keys and values must
be `bytes` objects (so, if you want to store objects of other types, you
need to serialize them to `bytes` first).

The module is based on the well-known BerkelyDB library, version 1.xx.

Example::

    import btree

    # First, we need to open a stream which holds a database
    # This is usually a file, but can be in-memory database
    # using io.BytesIO, a raw flash partition, etc.
    # Oftentimes, you want to create a database file if it doesn't
    # exist and open if it exists. Idiom below takes care of this.
    # DO NOT open database with "a+b" access mode.
    try:
        f = open("mydb", "r+b")
    except OSError:
        f = open("mydb", "w+b")

    # Now open a database itself
    db = btree.open(f)

    # The keys you add will be sorted internally in the database
    db[b"3"] = b"three"
    db[b"1"] = b"one"
    db[b"2"] = b"two"

    # Assume that any changes are cached in memory unless
    # explicitly flushed (or database closed). Flush database
    # at the end of each "transaction".
    db.flush()

    # Prints b'two'
    print(db[b"2"])

    # Iterate over sorted keys in the database, starting from b"2"
    # until the end of the database, returning only values.
    # Mind that arguments passed to values() method are *key* values.
    # Prints:
    #   b'two'
    #   b'three'
    for word in db.values(b"2"):
        print(word)

    del db[b"2"]

    # No longer true, prints False
    print(b"2" in db)

    # Prints:
    #  b"1"
    #  b"3"
    for key in db:
        print(key)

    db.close()

    # Don't forget to close the underlying stream!
    f.close()
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Any, Final, Iterable

from uio import IOBase

def open(
    stream: IOBase[bytes, Any],
    /,
    *,
    flags: int = 0,
    pagesize: int = 0,
    cachesize: int = 0,
    minkeypage: int = 0,
) -> _BTree:
    """
   Open a database from a random-access `stream` (like an open file). All
   other parameters are optional and keyword-only, and allow to tweak advanced
   parameters of the database operation (most users will not need them):
   
   * *flags* - Currently unused.
   * *pagesize* - Page size used for the nodes in BTree. Acceptable range
     is 512-65536. If 0, a port-specific default will be used, optimized for
     port's memory usage and/or performance.
   * *cachesize* - Suggested memory cache size in bytes. For a
     board with enough memory using larger values may improve performance.
     Cache policy is as follows: entire cache is not allocated at once;
     instead, accessing a new page in database will allocate a memory buffer
     for it, until value specified by *cachesize* is reached. Then, these
     buffers will be managed using LRU (least recently used) policy. More
     buffers may still be allocated if needed (e.g., if a database contains
     big keys and/or values). Allocated cache buffers aren't reclaimed.
   * *minkeypage* - Minimum number of keys to store per page. Default value
     of 0 equivalent to 2.
   
   Returns a BTree object, which implements a dictionary protocol (set
   of methods), and some additional methods described below.
   """

INCL: Final[int] = ...
"""
A flag for `keys()`, `values()`, `items()` methods to specify that
   scanning should be inclusive of the end key.
"""

DESC: Final[int] = ...
"""
A flag for `keys()`, `values()`, `items()` methods to specify that
   scanning should be in descending direction of keys.
"""

class _BTree:
    """

   """

    def close(self) -> None:
        """
      Close the database. It's mandatory to close the database at the end of
      processing, as some unwritten data may be still in the cache. Note that
      this does not close underlying stream with which the database was opened,
      it should be closed separately (which is also mandatory to make sure that
      data flushed from buffer to the underlying storage).
      """
    def flush(self) -> None:
        """
      Flush any data in cache to the underlying stream.
      """
    def __getitem__(self, key: bytes, /) -> bytes:
        """
      Standard dictionary methods.
      """
    def get(self, key: bytes, default: bytes | None = None, /) -> bytes | None:
        """
      Standard dictionary methods.
      """
    def __setitem__(self, key: bytes, val: bytes, /) -> None:
        """
      Standard dictionary methods.
      """
    def __delitem__(self, key: bytes, /) -> None:
        """
      Standard dictionary methods.
      """
    def __contains__(self, key: bytes, /) -> bool:
        """
      Standard dictionary methods.
      """
    def __iter__(self) -> Iterable[bytes]:
        """
      A BTree object can be iterated over directly (similar to a dictionary)
      to get access to all keys in order.
      """
    def keys(
        self,
        start_key: bytes | None = None,
        end_key: bytes | None = None,
        flags: int = 0,
        /,
    ) -> Iterable[bytes]:
        """
      These methods are similar to standard dictionary methods, but also can
      take optional parameters to iterate over a key sub-range, instead of
      the entire database. Note that for all 3 methods, *start_key* and
      *end_key* arguments represent key values. For example, `values()`
      method will iterate over values corresponding to they key range
      given. None values for *start_key* means "from the first key", no
      *end_key* or its value of None means "until the end of database".
      By default, range is inclusive of *start_key* and exclusive of
      *end_key*, you can include *end_key* in iteration by passing *flags*
      of `btree.INCL`. You can iterate in descending key direction
      by passing *flags* of `btree.DESC`. The flags values can be ORed
      together.
      """
    def values(
        self,
        start_key: bytes | None = None,
        end_key: bytes | None = None,
        flags: int = 0,
        /,
    ) -> Iterable[bytes]:
        """
      These methods are similar to standard dictionary methods, but also can
      take optional parameters to iterate over a key sub-range, instead of
      the entire database. Note that for all 3 methods, *start_key* and
      *end_key* arguments represent key values. For example, `values()`
      method will iterate over values corresponding to they key range
      given. None values for *start_key* means "from the first key", no
      *end_key* or its value of None means "until the end of database".
      By default, range is inclusive of *start_key* and exclusive of
      *end_key*, you can include *end_key* in iteration by passing *flags*
      of `btree.INCL`. You can iterate in descending key direction
      by passing *flags* of `btree.DESC`. The flags values can be ORed
      together.
      """
    def items(
        self,
        start_key: bytes | None = None,
        end_key: bytes | None = None,
        flags: int = 0,
        /,
    ) -> Iterable[tuple[bytes, bytes]]:
        """
      These methods are similar to standard dictionary methods, but also can
      take optional parameters to iterate over a key sub-range, instead of
      the entire database. Note that for all 3 methods, *start_key* and
      *end_key* arguments represent key values. For example, `values()`
      method will iterate over values corresponding to they key range
      given. None values for *start_key* means "from the first key", no
      *end_key* or its value of None means "until the end of database".
      By default, range is inclusive of *start_key* and exclusive of
      *end_key*, you can include *end_key* in iteration by passing *flags*
      of `btree.INCL`. You can iterate in descending key direction
      by passing *flags* of `btree.DESC`. The flags values can be ORed
      together.
      """



import platform

from turcar import get_workbench
from turcar.languages import tr
from turcar.plugins.micropython.mp_front import (
    BareMetalMicroPythonConfigPage,
    BareMetalMicroPythonProxy,
    GenericBareMetalMicroPythonConfigPage,
    GenericBareMetalMicroPythonProxy,
    LocalMicroPythonConfigPage,
    LocalMicroPythonProxy,
    MicroPythonProxy,
    SshMicroPythonConfigPage,
    SshMicroPythonProxy,
    add_micropython_backend,
    list_serial_ports,
)


def load_plugin():
    add_micropython_backend(
        "GenericMicroPython",
        GenericBareMetalMicroPythonProxy,
        tr("MicroPython (generic)"),
        GenericBareMetalMicroPythonConfigPage,
        sort_key="49",
    )

    get_workbench().set_default("serial.last_backend_per_vid_pid", {})

    if platform.system() in ("Linux", "Darwin"):
        add_micropython_backend(
            "LocalMicroPython",
            LocalMicroPythonProxy,
            tr("MicroPython (local)"),
            LocalMicroPythonConfigPage,
            bare_metal=False,
            local_rtc=False,
            sort_key="21",
        )
        get_workbench().set_default("LocalMicroPython.executable", "micropython")

    add_micropython_backend(
        "SshMicroPython",
        SshMicroPythonProxy,
        tr("MicroPython (SSH)"),
        SshMicroPythonConfigPage,
        bare_metal=False,
        local_rtc=False,
        sort_key="22",
    )
    get_workbench().set_default("SshMicroPython.executable", "micropython")
    get_workbench().set_default("SshMicroPython.cwd", None)
    get_workbench().set_default("SshMicroPython.host", "")
    get_workbench().set_default("SshMicroPython.user", "")
    get_workbench().set_default("SshMicroPython.auth_method", "password")
    get_workbench().set_default("SshMicroPython.make_uploaded_shebang_scripts_executable", True)

    get_workbench().set_default("esptool.show_advanced_options", False)



from enum import Enum


class Port(Enum):
    """
    Port on the programmable brick or hub.

    Note:
        Motor Ports: A, B, C, D
        Sensor Ports: S1, S2, S3, S4
    """
    A = 'A'
    B = 'B'
    C = 'C'
    D = 'D'
    S1 = 'S1'
    S2 = 'S2'
    S3 = 'S3'
    S4 = 'S4'



"""mathematical functions

The `math` module provides some basic mathematical functions for
working with floating-point numbers.

|see_cpython_module| :mod:`cpython:math`.
"""

from __future__ import annotations

from typing import Tuple

e: float
"""base of the natural logarithm"""

pi: float
"""the ratio of a circle's circumference to its diameter"""

def acos(x: float) -> float:
    """Return the inverse cosine of ``x``."""
    ...

def asin(x: float) -> float:
    """Return the inverse sine of ``x``."""
    ...

def atan(x: float) -> float:
    """Return the inverse tangent of ``x``."""
    ...

def atan2(y: float, x: float) -> float:
    """Return the principal value of the inverse tangent of ``y/x``."""
    ...

def ceil(x: float) -> int:
    """Return an integer, being ``x`` rounded towards positive infinity."""
    ...

def copysign(x: float, y: float) -> float:
    """Return ``x`` with the sign of ``y``."""
    ...

def cos(x: float) -> float:
    """Return the cosine of ``x``."""
    ...

def degrees(x: float) -> float:
    """Return radians ``x`` converted to degrees."""
    ...

def exp(x: float) -> float:
    """Return the exponential of ``x``."""
    ...

def fabs(x: float) -> float:
    """Return the absolute value of ``x``."""
    ...

def floor(x: float) -> int:
    """Return an integer, being ``x`` rounded towards negative infinity."""
    ...

def fmod(x: float, y: float) -> int:
    """Return the remainder of ``x/y``."""
    ...

def frexp(x: float) -> Tuple[int, int]:
    """Decomposes a floating-point number into its mantissa and exponent.
    The returned value is the tuple ``(m, e)`` such that ``x == m * 2**e``
    exactly.  If ``x == 0`` then the function returns ``(0.0, 0)``, otherwise
    the relation ``0.5 <= abs(m) < 1`` holds."""
    ...

def isfinite(x: float) -> bool:
    """Return ``True`` if ``x`` is finite."""
    ...

def isinf(x: float) -> bool:
    """Return ``True`` if ``x`` is infinite."""
    ...

def isnan(x: float) -> bool:
    """Return ``True`` if ``x`` is not-a-number"""
    ...

def ldexp(x: float, exp: float) -> float:
    """Return ``x * (2**exp)``."""
    ...

def log(x: float, base: float = e) -> float:
    """Return the logarithm of x to the given base. If base is not specified,
    returns the natural logarithm (base e) of x"""
    ...

def modf(x: float) -> Tuple[float, float]:
    """Return a tuple of two floats, being the fractional and integral parts of
    ``x``.  Both return values have the same sign as ``x``."""
    ...

def pow(x: float, y: float) -> float:
    """Returns ``x`` to the power of ``y``."""

def radians(x: float) -> float:
    """Return degrees ``x`` converted to radians."""

def sin(x: float) -> float:
    """Return the sine of ``x``."""
    ...

def sqrt(x: float) -> float:
    """Returns the square root of ``x``."""
    ...

def tan(x: float) -> float:
    """Return the tangent of ``x``."""
    ...

def trunc(x: float) -> int:
    """Return an integer, being ``x`` rounded towards 0."""
    ...

def expm1(x: float) -> float:
    """Return ``exp(x) - 1``.

    May not be available on some boards.
    """
    ...

def log2(x: float) -> float:
    """Return the base-2 logarithm of ``x``.

    May not be available on some boards.
    """
    ...

def log10(x: float) -> float:
    """Return the base-10 logarithm of ``x``.

    May not be available on some boards.
    """
    ...

def cosh(x: float) -> float:
    """Return the hyperbolic cosine of ``x``.

    May not be available on some boards.
    """
    ...

def sinh(x: float) -> float:
    """Return the hyperbolic sine of ``x``.

    May not be available on some boards.
    """
    ...

def tanh(x: float) -> float:
    """Return the hyperbolic tangent of ``x``.

    May not be available on some boards.
    """
    ...

def acosh(x: float) -> float:
    """Return the inverse hyperbolic cosine of ``x``.

    May not be available on some boards.
    """
    ...

def asinh(x: float) -> float:
    """Return the inverse hyperbolic sine of ``x``.

    May not be available on some boards.
    """
    ...

def atanh(x: float) -> float:
    """Return the inverse hyperbolic tangent of ``x``.

    May not be available on some boards.
    """
    ...

def erf(x: float) -> float:
    """Return the error function of ``x``.

    May not be available on some boards.
    """
    ...

def erfc(x: float) -> float:
    """Return the complementary error function of ``x``.

    May not be available on some boards.
    """
    ...

def gamma(x: float) -> float:
    """Return the gamma function of ``x``.

    May not be available on some boards.
    """
    ...

def lgamma(x: float) -> float:
    """Return the natural logarithm of the gamma function of ``x``.

    May not be available on some boards.
    """
    ...



import re
import sys
import subprocess
import tkinter as tk
import os
import shutil
import threading


class CheackUpdates:
    def __init__(self):
        self.os_type = sys.platform
        self.root = tk.Tk()
        self.url = f"https://pypi.org/pypi/turcar/json"

    def is_windows(self):
        return self.os_type.startswith('win')

    def is_linux(self):
        return self.os_type.startswith('linux')

    def is_mac(self):
        return self.os_type.startswith('darwin')

    def close(self):
        self.root.destroy()

    # 检查版本
    def cheack_version(self, version):
        self.root.title('新版本更新')
        self.root.resizable(False, False)
        self.root.geometry("300x200")  # 设置弹框大小
        popup_label = tk.Label(self.root, text="发现新版本，是否立即更新？", font=("Arial", 16))
        popup_label.pack(pady=20)

        self.update_text = tk.StringVar()
        self.update_text.set("更新")
        self.popup_button_update = tk.Button(self.root, textvariable=self.update_text,
                                             command=lambda: self.update_bash('/home/orangepi/Downloads/turcar.bash',
                                                                              version))
        self.popup_button_update.pack(pady=10)

        popup_button_not_update = tk.Button(self.root, text="不更新", command=self.close)
        popup_button_not_update.pack(pady=10)
        self.root.geometry(
            "500x200+{}+{}".format(int(self.root.winfo_screenwidth() / 2 - 250),
                                   int(self.root.winfo_screenheight() / 2 - 100)))  # 根据窗口大小调整位置
        self.root.mainloop()

    def update_text_fun(self):
        self.update_text.set('正在拼命更新中，更新完毕自动重启，然后就能使用心爱的turcar啦')

    # 更新脚本中的版本号
    def update_bash(self, filename, new_version):
        # from turcar import workbench
        # bench = workbench.Workbench()
        # return bench._on_close()
        # 丢给多线程是因为直接写不更新，因为bash给他阻塞了，但是好像并没有什么作用
        threading.Thread(target=self.update_text_fun).start()

        """
        更新脚本中的版本号

        Args:
            filename: 要更新的脚本文件名
            new_version: 新的版本号，格式为 x.x.x
        """
        if os.path.exists(filename):
            pass
        else:
            # 使用示例
            source_file = "turcar-update.bash"
            destination_file = "/home/orangepi/Downloads/turcar.bash"
            copy_file(source_file, destination_file)

        # 使用正则表达式匹配并替换版本号
        with open(filename, 'r') as file:
            content = file.read()
            new_content = re.sub(r'(VERSION=)\d+\.\d+\.\d+', r'\g<1>' + new_version, content)

        # 将更新后的内容写回脚本文件
        with open(filename, 'w') as file:
            file.write(new_content)

        # 指定版本更新

        # thread = threading.Thread(target=execute_bash_command, args=(list(filename), ))
        #
        # # 开始线程
        # thread.start()
        #
        # # 等待线程完成
        # thread.join()
        self.popup_button_update.config(state=tk.DISABLED)
        self.update_text.set("更新中...")
        self.root.update()

        execute_bash_command(filename)

        # 显示更新完成的提示消息，关闭窗口
        self.update_text.set("更新完成")
        self.root.after(2000, self.close)  # 2秒后关闭窗口
        self.root.destroy()



def copy_file(source_file, destination_file):
    try:
        shutil.copy(source_file, destination_file)
        print("文件复制成功！")
    except FileNotFoundError:
        print("找不到源文件，请检查路径是否正确。")


# 执行 Bash 命令
def execute_bash_command(command):
    result = subprocess.run(command)
    return True

    # 输出命令执行结果
    # print("命令输出：")
    # print(result.stdout)
    # print()
    #
    # # 输出命令的退出状态码
    # print("退出状态码：", result.returncode)



"""
functionality specific to the ESP32.

Descriptions taken from:
https://raw.githubusercontent.com/micropython/micropython/master/docs/library/esp32.rst.
====================================================

.. module:: esp32
    :synopsis: functionality specific to the ESP32

The ``esp32`` module contains functions and classes specifically aimed at
controlling ESP32 modules.
"""

__author__ = "Howard C Lovatt"
__copyright__ = "Howard C Lovatt, 2020 onwards."
__license__ = "MIT https://opensource.org/licenses/MIT (as used by MicroPython)."
__version__ = "7.3.9"  # Version set by https://github.com/hlovatt/tag2ver

from typing import Final, List, Tuple, overload, ClassVar

from machine import Pin
from uio import AnyReadableBuf
from uos import AbstractBlockDev

def wake_on_touch(wake: bool, /) -> None:
    """
    Configure whether or not a touch will wake the device from sleep.
    *wake* should be a boolean value.
   """

def wake_on_ext0(pin: Pin | None, level: int, /) -> None:
    """
    Configure how EXT0 wakes the device from sleep.  *pin* can be ``None``
    or a valid Pin object.  *level* should be ``esp32.WAKEUP_ALL_LOW`` or
    ``esp32.WAKEUP_ANY_HIGH``.
   """

def wake_on_ext1(pins: List[Pin] | Tuple[Pin, ...] | None, level: int, /) -> None:
    """
    Configure how EXT1 wakes the device from sleep.  *pins* can be ``None``
    or a tuple/list of valid Pin objects.  *level* should be ``esp32.WAKEUP_ALL_LOW``
    or ``esp32.WAKEUP_ANY_HIGH``.
   """

def raw_temperature() -> int:
    """
    Read the raw value of the internal temperature sensor, returning an integer.
   """

def hall_sensor() -> int:
    """
    Read the raw value of the internal Hall sensor, returning an integer.
   """

def idf_heap_info(capabilities: int) -> List[Tuple[int, int, int, int]]:
    """
    Returns information about the ESP-IDF heap memory regions. One of them contains
    the MicroPython heap and the others are used by ESP-IDF, e.g., for network
    buffers and other data. This data is useful to get a sense of how much memory
    is available to ESP-IDF and the networking stack in particular. It may shed
    some light on situations where ESP-IDF operations fail due to allocation failures.
    The information returned is *not* useful to troubleshoot Python allocation failures,
    use `micropython.mem_info()` instead.
    
    The capabilities parameter corresponds to ESP-IDF's ``MALLOC_CAP_XXX`` values but the
    two most useful ones are predefined as `esp32.HEAP_DATA` for data heap regions and
    `esp32.HEAP_EXEC` for executable regions as used by the native code emitter.
    
    The return value is a list of 4-tuples, where each 4-tuple corresponds to one heap
    and contains: the total bytes, the free bytes, the largest free block, and
    the minimum free seen over time.
    
    Example after booting::
    
        >>> import esp32; esp32.idf_heap_info(esp32.HEAP_DATA)
        [(240, 0, 0, 0), (7288, 0, 0, 0), (16648, 4, 4, 4), (79912, 35712, 35512, 35108),
         (15072, 15036, 15036, 15036), (113840, 0, 0, 0)]
   """

HEAP_DATA: Final[int] = ...
"""
Used in `idf_heap_info`.
"""

HEAP_EXEC: Final[int] = ...
"""
Used in `idf_heap_info`.
"""

WAKEUP_ALL_LOW: Final[int] = ...
"""
Selects the wake level for pins.
"""

WAKEUP_ANY_HIGH: Final[int] = ...
"""
Selects the wake level for pins.
"""

class Partition(AbstractBlockDev):
    """
   This class gives access to the partitions in the device's flash memory and includes
   methods to enable over-the-air (OTA) updates.
   """

    BOOT: ClassVar[int] = ...
    """
Used in the `Partition` constructor to fetch various partitions: ``BOOT`` is the
    partition that will be booted at the next reset and ``RUNNING`` is the currently
    running partition.
   """

    RUNNING: ClassVar[int] = ...
    """
Used in the `Partition` constructor to fetch various partitions: ``BOOT`` is the
    partition that will be booted at the next reset and ``RUNNING`` is the currently
    running partition.
   """

    TYPE_APP: ClassVar[int] = ...
    """
Used in `Partition.find` to specify the partition type: ``APP`` is for bootable
    firmware partitions (typically labelled ``factory``, ``ota_0``, ``ota_1``), and
    ``DATA`` is for other partitions, e.g. ``nvs``, ``otadata``, ``phy_init``, ``vfs``.
   """

    TYPE_DATA: ClassVar[int] = ...
    """
Used in `Partition.find` to specify the partition type: ``APP`` is for bootable
    firmware partitions (typically labelled ``factory``, ``ota_0``, ``ota_1``), and
    ``DATA`` is for other partitions, e.g. ``nvs``, ``otadata``, ``phy_init``, ``vfs``.
   """
    def __init__(self, id: str | int, /):
        """
       Create an object representing a partition.  *id* can be a string which is the label
       of the partition to retrieve, or one of the constants: ``BOOT`` or ``RUNNING``.
      """
    @staticmethod
    def find(
        type: int = TYPE_APP, subtype: int = 0xFF, label: str | None = None, /
    ) -> List["Partition"]:
        """
       Find a partition specified by *type*, *subtype* and *label*.  Returns a
       (possibly empty) list of Partition objects. Note: ``subtype=0xff`` matches any subtype
       and ``label=None`` matches any label.
      """
    def info(self) -> (int, int, int, int, str, bool):
        """
       Returns a 6-tuple ``(type, subtype, addr, size, label, encrypted)``.
      """
    def set_boot(self) -> None:
        """
       Sets the partition as the boot partition.
      """
    def get_next_update(self) -> "Partition":
        """
       Gets the next update partition after this one, and returns a new Partition object.
       Typical usage is ``Partition(Partition.RUNNING).get_next_update()``
       which returns the next partition to update given the current running one.
      """
    @staticmethod
    def mark_app_valid_cancel_rollback() -> None:
        """
       Signals that the current boot is considered successful.
       Calling ``mark_app_valid_cancel_rollback`` is required on the first boot of a new
       partition to avoid an automatic rollback at the next boot.
       This uses the ESP-IDF "app rollback" feature with "CONFIG_BOOTLOADER_APP_ROLLBACK_ENABLE"
       and  an ``OSError(-261)`` is raised if called on firmware that doesn't have the
       feature enabled.
       It is OK to call ``mark_app_valid_cancel_rollback`` on every boot and it is not
       necessary when booting firmare that was loaded using esptool.
      """

class RMT:
    """
   The RMT (Remote Control) module, specific to the ESP32, was originally designed
   to send and receive infrared remote control signals. However, due to a flexible
   design and very accurate (as low as 12.5ns) pulse generation, it can also be
   used to transmit or receive many other types of digital signals::
   
       import esp32
       from machine import Pin
   
       r = esp32.RMT(0, pin=Pin(18), clock_div=8)
       r  # RMT(channel=0, pin=18, source_freq=80000000, clock_div=8, idle_level=0)
   
       # To apply a carrier frequency to the high output
       r = esp32.RMT(0, pin=Pin(18), clock_div=8, tx_carrier=(38000, 50, 1))
   
       # The channel resolution is 100ns (1/(source_freq/clock_div)).
       r.write_pulses((1, 20, 2, 40), 0)  # Send 0 for 100ns, 1 for 2000ns, 0 for 200ns, 1 for 4000ns
   
   The input to the RMT module is an 80MHz clock (in the future it may be able to
   configure the input clock but, for now, it's fixed). ``clock_div`` *divides*
   the clock input which determines the resolution of the RMT channel. The
   numbers specified in ``write_pulses`` are multiplied by the resolution to
   define the pulses.
   
   ``clock_div`` is an 8-bit divider (0-255) and each pulse can be defined by
   multiplying the resolution by a 15-bit (0-32,768) number. There are eight
   channels (0-7) and each can have a different clock divider.
   
   So, in the example above, the 80MHz clock is divided by 8. Thus the
   resolution is (1/(80Mhz/8)) 100ns. Since the ``start`` level is 0 and toggles
   with each number, the bitstream is ``0101`` with durations of [100ns, 2000ns,
   100ns, 4000ns].
   
   For more details see Espressif's `ESP-IDF RMT documentation.
   <https://docs.espressif.com/projects/esp-idf/en/latest/api-reference/peripherals/rmt.html>`_.
   
   .. Warning::
      The current MicroPython RMT implementation lacks some features, most notably
      receiving pulses. RMT should be considered a
      *beta feature* and the interface may change in the future.
   """

    def __init__(
        self,
        channel: int,
        /,
        *,
        pin: Pin | None = None,
        clock_div: int = 8,
        idle_level: bool = False,
        tx_carrier: Tuple[int, int, bool] | None = None,
    ):
        """
       This class provides access to one of the eight RMT channels. *channel* is
       required and identifies which RMT channel (0-7) will be configured. *pin*,
       also required, configures which Pin is bound to the RMT channel. *clock_div*
       is an 8-bit clock divider that divides the source clock (80MHz) to the RMT
       channel allowing the resolution to be specified. *idle_level* specifies
       what level the output will be when no transmission is in progress and can
       be any value that converts to a boolean, with ``True`` representing high
       voltage and ``False`` representing low.
       
       To enable the transmission carrier feature, *tx_carrier* should be a tuple
       of three positive integers: carrier frequency, duty percent (``0`` to
       ``100``) and the output level to apply the carrier to (a boolean as per
       *idle_level*).
      """
    def source_freq(self) -> int:
        """
       Returns the source clock frequency. Currently the source clock is not
       configurable so this will always return 80MHz.
      """
    def clock_div(self) -> int:
        """
       Return the clock divider. Note that the channel resolution is
       ``1 / (source_freq / clock_div)``.
      """
    def wait_done(self, *, timeout: int = 0) -> bool:
        """
       Returns ``True`` if the channel is idle or ``False`` if a sequence of
       pulses started with `RMT.write_pulses` is being transmitted. If the
       *timeout* keyword argument is given then block for up to this many
       milliseconds for transmission to complete.
      """
    def loop(self, enable_loop: bool, /) -> None:
        """
       Configure looping on the channel. *enable_loop* is bool, set to ``True`` to
       enable looping on the *next* call to `RMT.write_pulses`. If called with
       ``False`` while a looping sequence is currently being transmitted then the
       current loop iteration will be completed and then transmission will stop.
      """
    @overload
    def write_pulses(
        self, duration: List[int] | Tuple[int, ...], data: bool = True, /
    ) -> None:
        """
       Begin transmitting a sequence. There are three ways to specify this:
       
       **Mode 1:** *duration* is a list or tuple of durations. The optional *data*
       argument specifies the initial output level. The output level will toggle
       after each duration.
       
       **Mode 2:** *duration* is a positive integer and *data* is a list or tuple
       of output levels. *duration* specifies a fixed duration for each.
       
       **Mode 3:** *duration* and *data* are lists or tuples of equal length,
       specifying individual durations and the output level for each.
       
       Durations are in integer units of the channel resolution (as described
       above), between 1 and 32767 units. Output levels are any value that can
       be converted to a boolean, with ``True`` representing high voltage and
       ``False`` representing low.
       
       If transmission of an earlier sequence is in progress then this method will
       block until that transmission is complete before beginning the new sequence.
       
       If looping has been enabled with `RMT.loop`, the sequence will be
       repeated indefinitely. Further calls to this method will block until the
       end of the current loop iteration before immediately beginning to loop the
       new sequence of pulses. Looping sequences longer than 126 pulses is not
       supported by the hardware.
      """
    @overload
    def write_pulses(
        self, duration: int, data: List[bool] | Tuple[bool, ...], /
    ) -> None:
        """
       Begin transmitting a sequence. There are three ways to specify this:
       
       **Mode 1:** *duration* is a list or tuple of durations. The optional *data*
       argument specifies the initial output level. The output level will toggle
       after each duration.
       
       **Mode 2:** *duration* is a positive integer and *data* is a list or tuple
       of output levels. *duration* specifies a fixed duration for each.
       
       **Mode 3:** *duration* and *data* are lists or tuples of equal length,
       specifying individual durations and the output level for each.
       
       Durations are in integer units of the channel resolution (as described
       above), between 1 and 32767 units. Output levels are any value that can
       be converted to a boolean, with ``True`` representing high voltage and
       ``False`` representing low.
       
       If transmission of an earlier sequence is in progress then this method will
       block until that transmission is complete before beginning the new sequence.
       
       If looping has been enabled with `RMT.loop`, the sequence will be
       repeated indefinitely. Further calls to this method will block until the
       end of the current loop iteration before immediately beginning to loop the
       new sequence of pulses. Looping sequences longer than 126 pulses is not
       supported by the hardware.
      """
    @overload
    def write_pulses(
        self,
        duration: List[int] | Tuple[int, ...],
        data: List[bool] | Tuple[bool, ...],
        /,
    ) -> None:
        """
       Begin transmitting a sequence. There are three ways to specify this:
       
       **Mode 1:** *duration* is a list or tuple of durations. The optional *data*
       argument specifies the initial output level. The output level will toggle
       after each duration.
       
       **Mode 2:** *duration* is a positive integer and *data* is a list or tuple
       of output levels. *duration* specifies a fixed duration for each.
       
       **Mode 3:** *duration* and *data* are lists or tuples of equal length,
       specifying individual durations and the output level for each.
       
       Durations are in integer units of the channel resolution (as described
       above), between 1 and 32767 units. Output levels are any value that can
       be converted to a boolean, with ``True`` representing high voltage and
       ``False`` representing low.
       
       If transmission of an earlier sequence is in progress then this method will
       block until that transmission is complete before beginning the new sequence.
       
       If looping has been enabled with `RMT.loop`, the sequence will be
       repeated indefinitely. Further calls to this method will block until the
       end of the current loop iteration before immediately beginning to loop the
       new sequence of pulses. Looping sequences longer than 126 pulses is not
       supported by the hardware.
      """

class ULP:
    """
   This class provides access to the Ultra-Low-Power co-processor.
   """

    def __init__(self):
        """
       This class provides access to the Ultra-Low-Power co-processor.
      """
    def set_wakeup_period(self, period_index: int, period_us: int, /) -> None:
        """
       Set the wake-up period.
      """
    def load_binary(self, load_addr: int, program_binary: AnyReadableBuf, /) -> None:
        """
       Load a *program_binary* into the ULP at the given *load_addr*.
      """
    def run(self, entry_point: int, /) -> None:
        """
       Start the ULP running at the given *entry_point*.
      """

class NVS:
    """
   This class gives access to the Non-Volatile storage managed by ESP-IDF. The NVS is partitioned
   into namespaces and each namespace contains typed key-value pairs. The keys are strings and the
   values may be various integer types, strings, and binary blobs. The driver currently only
   supports 32-bit signed integers and blobs.
   
   .. warning::
   
       Changes to NVS need to be committed to flash by calling the commit method. Failure
       to call commit results in changes being lost at the next reset.
   """

    def __init__(self, namespace: str, /):
        """
       Create an object providing access to a namespace (which is automatically created if not
       present).
      """
    def set_i32(self, key: str, value: int, /) -> None:
        """
       Sets a 32-bit signed integer value for the specified key. Remember to call *commit*!
      """
    def get_i32(self, key: str, /) -> int:
        """
       Returns the signed integer value for the specified key. Raises an OSError if the key does not
       exist or has a different type.
      """
    def set_blob(self, key: str, value: AnyReadableBuf, /) -> None:
        """
       Sets a binary blob value for the specified key. The value passed in must support the buffer
       protocol, e.g. bytes, bytearray, str. (Note that esp-idf distinguishes blobs and strings, this
       method always writes a blob even if a string is passed in as value.)
       Remember to call *commit*!
      """
    def get_blob(self, key: str, buffer: bytearray, /) -> int:
        """
       Reads the value of the blob for the specified key into the buffer, which must be a bytearray.
       Returns the actual length read. Raises an OSError if the key does not exist, has a different
       type, or if the buffer is too small.
      """
    def erase_key(self, key: str, /) -> None:
        """
       Erases a key-value pair.
      """
    def commit(self) -> None:
        """
       Commits changes made by *set_xxx* methods to flash.
      """



"""
/usr/bin/pybricks-repl content:
#!/bin/sh
brickrun -r -- pybricks-micropython -i -c 'from core import *'

help("modules"):

__main__          mmap              pybricks/tools    ufcntl
_thread           nxtdevices_c      pybricks/uev3dev/__init__           uhashlib
array             parameters_c      pybricks/uev3dev/_alsa              uheapq
bluetooth_c       pybricks/__init__ pybricks/uev3dev/_wand              uio
btree             pybricks/bluetooth                  pybricks/uev3dev/display            ujson
builtins          pybricks/display  pybricks/uev3dev/i2c                umachine
cmath             pybricks/ev3brick pybricks/uev3dev/messaging          uos
core              pybricks/ev3devices                 pybricks/uev3dev/sound              urandom
ev3devices_c      pybricks/ev3devio pybricks/uev3dev/util               ure
experimental_c    pybricks/experimental               robotics_c        uselect
ffi               pybricks/hubs     sys               usignal
framebuf          pybricks/iodevices                  termios           usocket
gc                pybricks/media/ev3dev               tools             ussl
hubs_c            pybricks/messaging                  ubinascii         ustruct
iodevices_c       pybricks/nxtdevices                 ucollections      utime
math              pybricks/parameters                 ucryptolib        utimeq
media_ev3dev_c    pybricks/robotics uctypes           uwebsocket
micropython       pybricks/speaker  uerrno            uzlib
Plus any modules on the filesystem


"""
from tkinter import ttk

from turcar import get_workbench
from turcar.languages import tr
from turcar.misc_utils import running_on_windows
from turcar.plugins.micropython import (
    SshMicroPythonConfigPage,
    SshMicroPythonProxy,
    add_micropython_backend,
)
from turcar.ui_utils import create_url_label, ems_to_pixels


class EV3MicroPythonProxy(SshMicroPythonProxy):
    def _get_launcher_with_args(self):
        import turcar.plugins.ev3.ev3_back

        args = {
            "cwd": get_workbench().get_option(f"{self.backend_name}.cwd") or "",
            "interpreter": self._target_executable,
            "host": self._host,
            "user": self._user,
        }

        args.update(self._get_time_args())
        args.update(self._get_extra_launcher_args())

        cmd = [
            turcar.plugins.ev3.ev3_back.__file__,
            repr(args),
        ]
        return cmd

    def _get_extra_launcher_args(self):
        return {"interpreter_launcher": ["brickrun", "-r", "--"]}


class EV3MicroPythonConfigPage(SshMicroPythonConfigPage):
    def __init__(self, master):
        super().__init__(master)
        inner_pad = ems_to_pixels(0.6)

        preps_label = ttk.Label(self, text=tr("Preparations (skip the VS Code part)"))
        preps_label.grid(row=0, column=0, pady=(0, inner_pad), sticky="w")

        pybricks_url = create_url_label(self, "https://pybricks.com/ev3-micropython/")
        pybricks_url.grid(row=0, column=1, pady=(0, inner_pad), padx=ems_to_pixels(1), sticky="w")

        ttk.Label(self, text=tr("Default password")).grid(
            row=6, column=0, pady=(0, inner_pad), sticky="w"
        )
        default_pw_box = ttk.Entry(self)
        default_pw_box.insert(0, "maker")
        default_pw_box["state"] = "disabled"
        default_pw_box.grid(
            row=6, column=1, pady=(0, inner_pad), padx=ems_to_pixels(1), sticky="we"
        )

    def has_editable_interpreter(self) -> bool:
        return False


def load_plugin():
    add_micropython_backend(
        "EV3MicroPython",
        EV3MicroPythonProxy,
        "MicroPython (EV3)",
        EV3MicroPythonConfigPage,
        bare_metal=False,
        sort_key="23",
    )
    get_workbench().set_default("EV3MicroPython.executable", "pybricks-micropython")
    get_workbench().set_default("EV3MicroPython.make_uploaded_shebang_scripts_executable", True)
    get_workbench().set_default("EV3MicroPython.cwd", None)
    get_workbench().set_default(
        "EV3MicroPython.host", "ev3dev" if running_on_windows() else "ev3dev.local"
    )
    get_workbench().set_default("EV3MicroPython.user", "robot")
    get_workbench().set_default("EV3MicroPython.auth_method", "password")



import os.path

from PIL import Image

d = os.path.dirname(__file__)

"""
for name in os.listdir(d):
    full_path = os.path.join(d, name)
    if full_path.endswith(".gif") and os.path.isfile(full_path.replace(".gif", ".png")):
        os.remove(full_path)
"""

os.chdir(d)
for name in os.listdir(d):
    if (name.endswith(".gif") or name.endswith(".png")) and "_disabled_" not in name:
        im = Image.open(name)
        trans_brightness = 160
        trans = Image.new(
            "RGBA", im.size, (trans_brightness, trans_brightness, trans_brightness, 0)
        )
        im = im.convert("RGBA")
        im = Image.blend(trans, im, 0.4)
        im.save("_disabled_" + name)



from __future__ import annotations

import contextlib
import logging
import os
import time
import warnings
from abc import ABC, abstractmethod
from threading import Lock
from types import TracebackType
from typing import Any

from ._error import Timeout

_LOGGER = logging.getLogger("filelock")


# This is a helper class which is returned by :meth:`BaseFileLock.acquire` and wraps the lock to make sure __enter__
# is not called twice when entering the with statement. If we would simply return *self*, the lock would be acquired
# again in the *__enter__* method of the BaseFileLock, but not released again automatically. issue #37 (memory leak)
class AcquireReturnProxy:
    """A context aware object that will release the lock file when exiting."""

    def __init__(self, lock: BaseFileLock) -> None:
        self.lock = lock

    def __enter__(self) -> BaseFileLock:
        return self.lock

    def __exit__(
        self,
        exc_type: type[BaseException] | None,  # noqa: U100
        exc_value: BaseException | None,  # noqa: U100
        traceback: TracebackType | None,  # noqa: U100
    ) -> None:
        self.lock.release()


class BaseFileLock(ABC, contextlib.ContextDecorator):
    """Abstract base class for a file lock object."""

    def __init__(self, lock_file: str | os.PathLike[Any], timeout: float = -1) -> None:
        """
        Create a new lock object.

        :param lock_file: path to the file
        :param timeout: default timeout when acquiring the lock, in seconds. It will be used as fallback value in
        the acquire method, if no timeout value (``None``) is given. If you want to disable the timeout, set it
        to a negative value. A timeout of 0 means, that there is exactly one attempt to acquire the file lock.
        """
        # The path to the lock file.
        self._lock_file: str = os.fspath(lock_file)

        # The file descriptor for the *_lock_file* as it is returned by the os.open() function.
        # This file lock is only NOT None, if the object currently holds the lock.
        self._lock_file_fd: int | None = None

        # The default timeout value.
        self._timeout: float = timeout

        # We use this lock primarily for the lock counter.
        self._thread_lock: Lock = Lock()

        # The lock counter is used for implementing the nested locking mechanism. Whenever the lock is acquired, the
        # counter is increased and the lock is only released, when this value is 0 again.
        self._lock_counter: int = 0

    @property
    def lock_file(self) -> str:
        """:return: path to the lock file"""
        return self._lock_file

    @property
    def timeout(self) -> float:
        """
        :return: the default timeout value, in seconds

        .. versionadded:: 2.0.0
        """
        return self._timeout

    @timeout.setter
    def timeout(self, value: float | str) -> None:
        """
        Change the default timeout value.

        :param value: the new value, in seconds
        """
        self._timeout = float(value)

    @abstractmethod
    def _acquire(self) -> None:
        """If the file lock could be acquired, self._lock_file_fd holds the file descriptor of the lock file."""
        raise NotImplementedError

    @abstractmethod
    def _release(self) -> None:
        """Releases the lock and sets self._lock_file_fd to None."""
        raise NotImplementedError

    @property
    def is_locked(self) -> bool:
        """

        :return: A boolean indicating if the lock file is holding the lock currently.

        .. versionchanged:: 2.0.0

            This was previously a method and is now a property.
        """
        return self._lock_file_fd is not None

    def acquire(
        self,
        timeout: float | None = None,
        poll_interval: float = 0.05,
        *,
        poll_intervall: float | None = None,
        blocking: bool = True,
    ) -> AcquireReturnProxy:
        """
        Try to acquire the file lock.

        :param timeout: maximum wait time for acquiring the lock, ``None`` means use the default :attr:`~timeout` is and
         if ``timeout < 0``, there is no timeout and this method will block until the lock could be acquired
        :param poll_interval: interval of trying to acquire the lock file
        :param poll_intervall: deprecated, kept for backwards compatibility, use ``poll_interval`` instead
        :param blocking: defaults to True. If False, function will return immediately if it cannot obtain a lock on the
         first attempt. Otherwise this method will block until the timeout expires or the lock is acquired.
        :raises Timeout: if fails to acquire lock within the timeout period
        :return: a context object that will unlock the file when the context is exited

        .. code-block:: python

            # You can use this method in the context manager (recommended)
            with lock.acquire():
                pass

            # Or use an equivalent try-finally construct:
            lock.acquire()
            try:
                pass
            finally:
                lock.release()

        .. versionchanged:: 2.0.0

            This method returns now a *proxy* object instead of *self*,
            so that it can be used in a with statement without side effects.

        """
        # Use the default timeout, if no timeout is provided.
        if timeout is None:
            timeout = self.timeout

        if poll_intervall is not None:
            msg = "use poll_interval instead of poll_intervall"
            warnings.warn(msg, DeprecationWarning, stacklevel=2)
            poll_interval = poll_intervall

        # Increment the number right at the beginning. We can still undo it, if something fails.
        with self._thread_lock:
            self._lock_counter += 1

        lock_id = id(self)
        lock_filename = self._lock_file
        start_time = time.monotonic()
        try:
            while True:
                with self._thread_lock:
                    if not self.is_locked:
                        _LOGGER.debug("Attempting to acquire lock %s on %s", lock_id, lock_filename)
                        self._acquire()

                if self.is_locked:
                    _LOGGER.debug("Lock %s acquired on %s", lock_id, lock_filename)
                    break
                elif blocking is False:
                    _LOGGER.debug("Failed to immediately acquire lock %s on %s", lock_id, lock_filename)
                    raise Timeout(self._lock_file)
                elif 0 <= timeout < time.monotonic() - start_time:
                    _LOGGER.debug("Timeout on acquiring lock %s on %s", lock_id, lock_filename)
                    raise Timeout(self._lock_file)
                else:
                    msg = "Lock %s not acquired on %s, waiting %s seconds ..."
                    _LOGGER.debug(msg, lock_id, lock_filename, poll_interval)
                    time.sleep(poll_interval)
        except BaseException:  # Something did go wrong, so decrement the counter.
            with self._thread_lock:
                self._lock_counter = max(0, self._lock_counter - 1)
            raise
        return AcquireReturnProxy(lock=self)

    def release(self, force: bool = False) -> None:
        """
        Releases the file lock. Please note, that the lock is only completely released, if the lock counter is 0. Also
        note, that the lock file itself is not automatically deleted.

        :param force: If true, the lock counter is ignored and the lock is released in every case/
        """
        with self._thread_lock:

            if self.is_locked:
                self._lock_counter -= 1

                if self._lock_counter == 0 or force:
                    lock_id, lock_filename = id(self), self._lock_file

                    _LOGGER.debug("Attempting to release lock %s on %s", lock_id, lock_filename)
                    self._release()
                    self._lock_counter = 0
                    _LOGGER.debug("Lock %s released on %s", lock_id, lock_filename)

    def __enter__(self) -> BaseFileLock:
        """
        Acquire the lock.

        :return: the lock object
        """
        self.acquire()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,  # noqa: U100
        exc_value: BaseException | None,  # noqa: U100
        traceback: TracebackType | None,  # noqa: U100
    ) -> None:
        """
        Release the lock.

        :param exc_type: the exception type if raised
        :param exc_value: the exception value if raised
        :param traceback: the exception traceback if raised
        """
        self.release()

    def __del__(self) -> None:
        """Called when the lock object is deleted."""
        self.release(force=True)


__all__ = [
    "BaseFileLock",
    "AcquireReturnProxy",
]



"""Support for Sharp Memory Display framebuffers

For more information about working with Sharp Memory Displays,
see `this Learn guide <https://learn.adafruit.com/adafruit-sharp-memory-display-breakout/circuitpython-displayio-setup>`_.
"""

from __future__ import annotations

import busio
import microcontroller

class SharpMemoryFramebuffer:
    """A framebuffer for a memory-in-pixel display. Sharp makes monochrome displays and JDI used
    to make 8-color displays.

    This initializes a display and connects it into CircuitPython. Unlike other
    objects in CircuitPython, Display objects live until `displayio.release_displays()`
    is called. This is done so that CircuitPython can use the display itself."""

    def __init__(
        self,
        spi_bus: busio.SPI,
        chip_select: microcontroller.Pin,
        width: int,
        height: int,
        baudrate: int = 2000000,
        jdi_display: bool = False,
    ) -> None:
        """Create a framebuffer for the memory-in-pixel display.

        :param busio.SPI spi_bus: The SPI bus that the display is connected to
        :param microcontroller.Pin chip_select: The pin connect to the display's chip select line
        :param int width: The width of the display in pixels
        :param int height: The height of the display in pixels
        :param int baudrate: The baudrate to communicate with the screen at
        :param bool jdi_display: When True, work with an 8-color JDI display. Otherwise, a monochrome Sharp display.
        """
        ...
    def deinit(self) -> None:
        """Free the resources (pins, timers, etc.) associated with this
        SharpMemoryFramebuffer instance.  After deinitialization, no further operations
        may be performed."""
        ...



"""Native framebuffer display driving

The `framebufferio` module contains classes to manage display output
including synchronizing with refresh rates and partial updating.
It is used in conjunction with classes from `displayio` to actually
place items on the display; and classes like `RGBMatrix` to actually
drive the display."""

from __future__ import annotations

import circuitpython_typing
import displayio
from circuitpython_typing import WriteableBuffer

class FramebufferDisplay:
    """Manage updating a display with framebuffer in RAM

    This initializes a display and connects it into CircuitPython. Unlike other
    objects in CircuitPython, Display objects live until `displayio.release_displays()`
    is called. This is done so that CircuitPython can use the display itself."""

    def __init__(
        self,
        framebuffer: circuitpython_typing.FrameBuffer,
        *,
        rotation: int = 0,
        auto_refresh: bool = True,
    ) -> None:
        """Create a Display object with the given framebuffer (a buffer, array, ulab.array, etc)

        :param ~circuitpython_typing.FrameBuffer framebuffer: The framebuffer that the display is connected to
        :param bool auto_refresh: Automatically refresh the screen
        :param int rotation: The rotation of the display in degrees clockwise. Must be in 90 degree increments (0, 90, 180, 270)
        """
        ...
    def show(self, group: displayio.Group) -> None:
        """Switches to displaying the given group of layers. When group is None, the default
        CircuitPython terminal will be shown.

        :param Group group: The group to show."""
        ...
    def refresh(
        self, *, target_frames_per_second: int = 60, minimum_frames_per_second: int = 1
    ) -> bool:
        """When auto refresh is off, waits for the target frame rate and then refreshes the display,
        returning True. If the call has taken too long since the last refresh call for the given
        target frame rate, then the refresh returns False immediately without updating the screen to
        hopefully help getting caught up.

        If the time since the last successful refresh is below the minimum frame rate, then an
        exception will be raised. Set minimum_frames_per_second to 0 to disable.

        When auto refresh is on, updates the display immediately. (The display will also update
        without calls to this.)

        :param int target_frames_per_second: How many times a second `refresh` should be called and the screen updated.
        :param int minimum_frames_per_second: The minimum number of times the screen should be updated per second.
        """
        ...
    auto_refresh: bool
    """True when the display is refreshed automatically."""
    brightness: float
    """The brightness of the display as a float. 0.0 is off and 1.0 is full brightness."""
    width: int
    """Gets the width of the framebuffer"""
    height: int
    """Gets the height of the framebuffer"""
    rotation: int
    """The rotation of the display as an int in degrees."""
    framebuffer: circuitpython_typing.FrameBuffer
    """The framebuffer being used by the display"""
    def fill_row(self, y: int, buffer: WriteableBuffer) -> WriteableBuffer:
        """Extract the pixels from a single row

        :param int y: The top edge of the area
        :param ~circuitpython_typing.WriteableBuffer buffer: The buffer in which to place the pixel data
        """
        ...
    root_group: displayio.Group
    """The root group on the display."""



from tkinter import ttk

from turcar import get_workbench
from turcar.config_ui import ConfigurationPage
from turcar.languages import tr


class RunDebugConfigurationPage(ConfigurationPage):
    def __init__(self, master):
        super().__init__(master)

        self.add_checkbox(
            "run.allow_running_unnamed_programs",
            tr("Allow running unnamed programs"),
            row=3,
            columnspan=3,
        )

        self.add_checkbox(
            "run.auto_cd",
            tr("Change working directory to script directory on Run / Debug"),
            row=5,
            columnspan=3,
        )
        self.add_checkbox(
            "run.warn_module_shadowing",
            tr("Warn if a user module shadows a library module"),
            row=6,
            columnspan=3,
            pady=(0, 20),
        )

        self.add_checkbox(
            "debugger.frames_in_separate_windows",
            tr("Show function calls (frames) in separate windows"),
            tooltip=tr("Uncheck if you want more traditional experience."),
            row=10,
            columnspan=3,
        )
        self.add_checkbox(
            "debugger.automatic_stack_view",
            tr("Open and close Stack view automatically"),
            tooltip=tr(
                "Opens the Stack view on first call and "
                + "closes it when program returns to main frame."
            ),
            row=20,
            columnspan=3,
        )
        self.add_checkbox(
            "debugger.allow_stepping_into_libraries",
            tr("Allow stepping into libraries (ie. outside of main script directory)"),
            tooltip=tr("May make debugging slower."),
            row=30,
            columnspan=3,
        )

        default_label = ttk.Label(self, text=tr("Preferred debugger"), anchor="w")
        default_label.grid(row=40, column=0, sticky="w", pady=(15, 0))
        self.add_combobox(
            "debugger.preferred_debugger",
            ["nicer", "faster", "birdseye"],
            width=8,
            row=40,
            column=1,
            padx=5,
            pady=(15, 0),
        )
        default_comment_label = ttk.Label(
            self, text=tr("(used when clicking Debug toolbar button)"), anchor="w"
        )
        default_comment_label.grid(row=40, column=2, sticky="w", pady=(15, 0))

        if get_workbench().get_option("run.birdseye_port", None):
            port_label = ttk.Label(self, text=tr("Birdseye port"), anchor="w")
            port_label.grid(row=50, column=0, sticky="w", pady=(5, 0))
            self.add_entry("run.birdseye_port", row=50, column=1, width=5, pady=(5, 0), padx=5)
            port_comment_label = ttk.Label(
                self, text=tr("(restart turcar after changing this)"), anchor="w"
            )
            port_comment_label.grid(row=50, column=2, sticky="w", pady=(5, 0))

        self.columnconfigure(2, weight=1)


def load_plugin():
    get_workbench().add_configuration_page("run", tr("Run & Debug"), RunDebugConfigurationPage, 50)



# import cv2
import platform

# 获取当前操作系统
current_os = platform.system()

# if current_os == "Windows":
#     save_path = "D:/copy.png"
# else:  # 默认为Ubuntu或其他Linux发行版
#     save_path = "/home/username/copy.png"
# cap = cv2.VideoCapture(0, cv2.CAP_DSHOW)  # 打开笔记本内置摄像头
# while cap.isOpened():  # 笔记本内置摄像头被打开后
#     ret, frame = cap.read()  # 从摄像头中实时读取视频
#     cv2.imshow("Video", frame)  # 在窗口中显示视频
#     k = cv2.waitKey(1)  # 图像的刷新时间为1毫秒
#     if k == ord(' '):  # 按下空格键
#         cv2.imshow('img', frame)  # 显示按下空格键时摄像头视频中的图像
#         cv2.imwrite(save_path, frame)  # 保存按下空格键时摄像头视频中的图像
#         cap.release()  # 关闭笔记本内置摄像头
#         cv2.destroyWindow("Video")  # 销毁名为Video的窗口
#         break
# cv2.destroyAllWindows()  # 销毁显示图像的窗口



"""ESP Ultra Low Power Processor Module

The `espulp` module adds ability to load and run
programs on the ESP32-Sx's ultra-low-power RISC-V processor.

.. code-block:: python

    import espulp
    import memorymap

    shared_mem = memorymap.AddressRange(start=0x50000000, length=1024)
    ulp = espulp.ULP()

    with open("program.bin", "rb") as f:
        program = f.read()

    ulp.run(program)
    print(shared_mem[0])
    # ulp.halt()
"""

from __future__ import annotations

from typing import Optional, Sequence

import microcontroller
from circuitpython_typing import ReadableBuffer

...

def get_rtc_gpio_number(pin: microcontroller.Pin) -> Optional[int]:
    """Return the RTC GPIO number of the given pin or None if not connected
    to RTC GPIO."""
    ...

class Architecture:
    """The ULP architectures available."""

    FSM: Architecture
    """The ULP Finite State Machine."""

    RISCV: Architecture
    """The ULP RISC-V Coprocessor."""

class ULP:
    def __init__(self, arch: Architecture = Architecture.FSM):
        """The ultra-low-power processor.

        Raises an exception if another ULP has been instantiated. This
        ensures that is is only used by one piece of code at a time.

        :param Architecture arch: The ulp arch"""
        ...
    def deinit(self) -> None:
        """Deinitialises the ULP and releases it for another program."""
        ...
    def __enter__(self) -> ULP:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def run(
        self, program: ReadableBuffer, *, pins: Sequence[microcontroller.Pin] = ()
    ) -> None:
        """Loads the program into ULP memory and then runs the program. The given pins are
           claimed and not reset until `halt()` is called.

        The program will continue to run even when the running Python is halted."""
        ...
    def halt(self) -> None:
        """Halts the running program and releases the pins given in `run()`."""
        ...
    arch: Architecture
    """The ulp architecture. (read-only)"""

class ULPAlarm:
    """Trigger an alarm when the ULP requests wake-up."""

    def __init__(self, ulp: ULP) -> None:
        """Create an alarm that will be triggered when the ULP requests wake-up.

        The alarm is not active until it is passed to an `alarm`-enabling function, such as
        `alarm.light_sleep_until_alarms()` or `alarm.exit_and_deep_sleep_until_alarms()`.

        :param ULP ulp: The ulp instance"""
        ...



import hashlib
import json
import os.path
import platform
import shlex
import shutil
import stat
import subprocess
import sys
import urllib.request
from dataclasses import dataclass
from logging import getLogger
from typing import Dict, List, Optional, Set, Tuple
from urllib.request import urlopen

import filelock
from filelock import BaseFileLock, FileLock

from pipkin.adapters import Adapter
from pipkin.common import UserError
from pipkin.proxy import start_proxy
from pipkin.util import (
    get_base_executable,
    get_user_cache_dir,
    get_venv_executable,
    get_venv_site_packages_path,
    parse_meta_dir_name,
)

logger = getLogger(__name__)

INITIAL_VENV_DISTS = ["pip", "setuptools", "pkg_resources", "wheel"]
INITIAL_VENV_FILES = ["easy_install.py"]
META_ENCODING = "utf-8"


@dataclass(frozen=True)
class DistInfo:
    key: str
    project_name: str
    version: str
    location: str


class Session:
    """
    Allows performing several commands in row without releasing the venv.
    """

    def __init__(self, adapter: Adapter, tty: bool = True):
        self._adapter = adapter
        self._venv_lock: Optional[BaseFileLock] = None
        self._venv_dir: Optional[str] = None
        self._quiet = False
        self._tty = tty

    def install(
        self,
        specs: Optional[List[str]] = None,
        requirement_files: Optional[List[str]] = None,
        constraint_files: Optional[List[str]] = None,
        pre: bool = False,
        no_deps: bool = False,
        no_mp_org: bool = False,
        index_url: Optional[str] = None,
        extra_index_urls: Optional[List[str]] = None,
        no_index: bool = False,
        find_links: Optional[str] = None,
        target: Optional[str] = None,
        user: bool = False,
        upgrade: bool = False,
        upgrade_strategy: str = "only-if-needed",
        force_reinstall: bool = False,
        compile: Optional[bool] = None,
        mpy_cross: Optional[str] = None,
        **_,
    ):
        logger.debug("Starting install")

        if compile is None and mpy_cross:
            compile = True

        args = ["install", "--no-compile", "--use-pep517"]

        if upgrade:
            args.append("--upgrade")
        if upgrade_strategy:
            args += ["--upgrade-strategy", upgrade_strategy]
        if force_reinstall:
            args.append("--force-reinstall")

        args += self._format_selection_args(
            specs=specs,
            requirement_files=requirement_files,
            constraint_files=constraint_files,
            pre=pre,
            no_deps=no_deps,
        )

        self._populate_venv()
        state_before = self._get_venv_state()
        self._invoke_pip_with_index_args(
            args,
            no_mp_org=no_mp_org,
            index_url=index_url,
            extra_index_urls=extra_index_urls or [],
            no_index=no_index,
            find_links=find_links,
        )
        state_after = self._get_venv_state()

        removed_meta_dirs = {name for name in state_before if name not in state_after}
        # removed meta dirs are expected when upgrading
        for meta_dir_name in removed_meta_dirs:
            self._report_progress(f"Removing {parse_meta_dir_name(meta_dir_name)[0]}")
            dist_name, _version = parse_meta_dir_name(meta_dir_name)
            self._adapter.remove_dist(dist_name)

        new_meta_dirs = {name for name in state_after if name not in state_before}
        changed_meta_dirs = {
            name
            for name in state_after
            if name in state_before and state_after[name] != state_before[name]
        }

        if new_meta_dirs or changed_meta_dirs:
            self._report_progress("Starting to apply changes to the target.")

        if target:
            effective_target = target
        elif user:
            effective_target = self._adapter.get_user_packages_path()
        else:
            effective_target = self._adapter.get_default_target()

        for meta_dir in changed_meta_dirs:
            self._report_progress(f"Removing old version of {parse_meta_dir_name(meta_dir)[0]}")
            # if target is specified by --target or --user, then don't touch anything
            # besides corresponding directory, regardless of the sys.path and possible hiding
            dist_name, _version = parse_meta_dir_name(meta_dir)
            if target:
                # pip doesn't remove old dist with --target unless --upgrade is given
                if upgrade:
                    self._adapter.remove_dist(dist_name=dist_name, target=target)
            elif user:
                self._adapter.remove_dist(
                    dist_name=dist_name, target=self._adapter.get_user_packages_path()
                )
            else:
                # remove the all installations of this dist, which would hide the new installation
                self._adapter.remove_dist(
                    dist_name=dist_name, target=effective_target, above_target=True
                )

        for meta_dir in new_meta_dirs | changed_meta_dirs:
            self._upload_dist_by_meta_dir(meta_dir, effective_target, compile, mpy_cross)

        if new_meta_dirs or changed_meta_dirs:
            self._report_progress("All changes applied.")

    def uninstall(
        self,
        packages: Optional[List[str]] = None,
        requirement_files: Optional[List[str]] = None,
        yes: bool = False,
        **_,
    ):
        args = ["uninstall", "--yes"]

        for rf in requirement_files or []:
            args += ["-r", rf]
        for package in packages or []:
            args.append(package)

        self._populate_venv()
        state_before = self._get_venv_state()
        self._invoke_pip(args)
        state_after = self._get_venv_state()

        removed_meta_dirs = {name for name in state_before if name not in state_after}
        if removed_meta_dirs:
            # NB! If you want to move confirmation back to pip process, then test the process
            # in Windows via turcar
            if not yes:
                names = [parse_meta_dir_name(d)[0] for d in removed_meta_dirs]
                if input(f"Proceed removing {', '.join(names)} (Y/n) at target? ").lower() == "n":
                    return

            self._report_progress("Starting to apply changes to the target.")

        for meta_dir_name in removed_meta_dirs:
            self._report_progress(f"Removing {parse_meta_dir_name(meta_dir_name)[0]}")
            dist_name, _version = parse_meta_dir_name(meta_dir_name)
            self._adapter.remove_dist(dist_name)

        if removed_meta_dirs:
            self._report_progress("All changes applied.")

    def list(
        self,
        outdated: bool = False,
        uptodate: bool = False,
        not_required: bool = False,
        pre: bool = False,
        paths: Optional[List[str]] = None,
        user: bool = False,
        format: str = "columns",
        no_mp_org: Optional[bool] = False,
        index_url: Optional[str] = None,
        extra_index_urls: Optional[List[str]] = None,
        no_index: bool = False,
        find_links: Optional[str] = None,
        excludes: Optional[List[str]] = None,
        **_,
    ):
        args = ["list"]

        if outdated:
            args.append("--outdated")
        if uptodate:
            args.append("--uptodate")
        if not_required:
            args.append("--not-required")
        if pre:
            args.append("--pre")
        if format:
            args += ["--format", format]

        args += self._format_exclusion_args(excludes)

        self._populate_venv(paths=paths, user=user)

        self._invoke_pip_with_index_args(
            args,
            no_mp_org=no_mp_org,
            index_url=index_url,
            extra_index_urls=extra_index_urls,
            no_index=no_index,
            find_links=find_links,
        )

    def basic_list(self) -> Set[DistInfo]:
        """
        Allows listing without requiring the venv.
        """
        dists_by_name = self._adapter.list_dists()
        result = set()
        for name in dists_by_name:
            meta_dir_name, location = dists_by_name[name]
            name, version = parse_meta_dir_name(meta_dir_name)
            result.add(DistInfo(key=name, project_name=name, version=version, location=location))

        return result

    def show(self, packages: List[str], **_):
        self._populate_venv()
        self._invoke_pip(["show"] + packages)

    def freeze(
        self,
        paths: Optional[List[str]] = None,
        user: bool = False,
        excludes: Optional[List[str]] = None,
        **_,
    ):
        args = ["freeze"]

        args += self._format_exclusion_args(excludes)

        self._populate_venv(paths=paths, user=user)
        self._invoke_pip(args)

    def check(self, **_):
        self._populate_venv()
        self._invoke_pip(["check"])

    def download(
        self,
        specs: Optional[List[str]] = None,
        requirement_files: Optional[List[str]] = None,
        constraint_files: Optional[List[str]] = None,
        pre: bool = False,
        no_deps: bool = False,
        no_mp_org: bool = False,
        index_url: Optional[str] = None,
        extra_index_urls: Optional[List[str]] = None,
        no_index: bool = False,
        find_links: Optional[str] = None,
        dest: Optional[str] = None,
        **_,
    ):
        args = ["download"]

        if dest:
            args += ["--dest", dest]

        args += self._format_selection_args(
            specs=specs,
            requirement_files=requirement_files,
            constraint_files=constraint_files,
            pre=pre,
            no_deps=no_deps,
        )

        self._populate_venv()
        self._invoke_pip_with_index_args(
            args,
            no_mp_org=no_mp_org,
            index_url=index_url,
            extra_index_urls=extra_index_urls,
            no_index=no_index,
            find_links=find_links,
        )

    def wheel(
        self,
        specs: Optional[List[str]] = None,
        requirement_files: Optional[List[str]] = None,
        constraint_files: Optional[List[str]] = None,
        pre: bool = False,
        no_deps: bool = False,
        no_mp_org: bool = False,
        index_url: Optional[str] = None,
        extra_index_urls: Optional[List[str]] = None,
        no_index: bool = False,
        find_links: Optional[str] = None,
        wheel_dir: Optional[str] = None,
        **_,
    ):
        args = ["wheel"]

        if wheel_dir:
            args += ["--wheel-dir", wheel_dir]

        args += self._format_selection_args(
            specs=specs,
            requirement_files=requirement_files,
            constraint_files=constraint_files,
            pre=pre,
            no_deps=no_deps,
        )

        self._populate_venv()
        self._invoke_pip_with_index_args(
            args,
            no_mp_org=no_mp_org,
            index_url=index_url,
            extra_index_urls=extra_index_urls,
            no_index=no_index,
            find_links=find_links,
        )

    def cache(self, cache_command: str, **_) -> None:
        if cache_command == "purge":
            if os.path.exists(self._get_pipkin_cache_dir()):
                shutil.rmtree(self._get_pipkin_cache_dir())
        elif not os.path.exists(self._get_pipkin_cache_dir()):
            print(f"Cache dir ({self._get_pipkin_cache_dir()}) not created yet")
        elif cache_command == "dir":
            print(self._get_pipkin_cache_dir())
        else:
            self._ensure_venv()
            self._invoke_pip(["cache", cache_command])

    def close(self) -> None:
        if self._venv_lock is not None:
            # self._clear_venv()
            self._venv_lock.release()

    def _format_exclusion_args(self, excludes: Optional[List[str]]) -> List[str]:
        args = []
        for exclude in (excludes or []) + ["pip", "pkg_resources", "setuptools", "wheel"]:
            args += ["--exclude", exclude]

        return args

    def _format_selection_args(
        self,
        specs: Optional[List[str]],
        requirement_files: Optional[List[str]],
        constraint_files: Optional[List[str]],
        pre: bool,
        no_deps: bool,
    ):
        args = []

        for path in requirement_files or []:
            args += ["-r", path]
        for path in constraint_files or []:
            args += ["-c", path]

        if no_deps:
            args.append("--no-deps")
        if pre:
            args.append("--pre")

        args += specs or []

        return args

    def _upload_dist_by_meta_dir(
        self, meta_dir_name: str, target: str, compile: bool, mpy_cross: Optional[str]
    ) -> None:
        self._report_progress(f"Copying {parse_meta_dir_name(meta_dir_name)[0]}", end="")
        rel_record_path = os.path.join(meta_dir_name, "RECORD")
        record_path = os.path.join(self._get_venv_site_packages_path(), rel_record_path)
        assert os.path.exists(record_path)

        target_record_lines = []

        with open(record_path, encoding=META_ENCODING) as fp:
            record_lines = fp.read().splitlines()

        for line in record_lines:
            rel_path = line.split(",")[0]
            # don't consider files installed to e.g. bin-directory
            if rel_path.startswith(".."):
                continue

            # don't consider absolute paths
            if os.path.isabs(rel_path):
                logger.warning("Skipping absolute path %s", rel_path)
                continue

            # only consider METADATA from meta dir
            if rel_path.startswith(meta_dir_name) and os.path.basename(rel_path) != "METADATA":
                continue

            full_path = os.path.normpath(
                os.path.join(self._get_venv_site_packages_path(), rel_path)
            )

            full_device_path = self._adapter.join_path(target, self._adapter.normpath(rel_path))

            if full_path.endswith(".py") and compile:
                self._compile_with_mpy_cross(
                    full_path, self._get_compiled_path(full_path), mpy_cross
                )
                # forget about the .py file
                full_path = self._get_compiled_path(full_path)
                full_device_path = self._get_compiled_path(full_device_path)
                rel_path = self._get_compiled_path(rel_path)

            with open(full_path, "rb") as source_fp:
                content = source_fp.read()

            if rel_path.startswith(meta_dir_name) and os.path.basename(rel_path) == "METADATA":
                content = self._trim_metadata(content)

            self._adapter.write_file(full_device_path, content)
            self._report_progress(".", end="")
            target_record_lines.append(self._adapter.normpath(rel_path) + ",,")

        # add RECORD (without hashes)
        target_record_lines.append(self._adapter.normpath(rel_record_path) + ",,")
        full_device_record_path = self._adapter.join_path(
            target, self._adapter.normpath(rel_record_path)
        )
        self._adapter.write_file(
            full_device_record_path, "\n".join(target_record_lines).encode(META_ENCODING)
        )

        # add linebreak for the report
        self._report_progress("")

    def _trim_metadata(self, content: bytes) -> bytes:
        # TODO:
        return content

    def _get_compiled_path(self, source_path: str) -> str:
        assert source_path.endswith(".py"), f"Source path: {source_path}"
        return source_path[: -len(".py")] + ".mpy"

    def _ensure_venv(self) -> None:
        if self._venv_lock is not None:
            return

        self._venv_lock, self._venv_dir = self._prepare_venv()

    def _prepare_venv(self) -> Tuple[BaseFileLock, str]:
        path = self._compute_venv_path()
        if not os.path.exists(path):
            self._report_progress("Preparing working environment ...")
            logger.info("Start preparing working environment at %s ...", path)
            venv_cmd = [
                sys.executable,
                "-I",
                "-m",
                "venv",
                path,
            ]
            subprocess.check_call(
                venv_cmd,
                executable=venv_cmd[0],
                stdin=subprocess.DEVNULL,
            )
            logger.info("Done creating venv")
            assert os.path.exists(path)
            pip_cmd = [
                get_venv_executable(path),
                "-I",
                "-m",
                "pip",
                "--disable-pip-version-check",
                "install",
                "--no-warn-script-location",
                "--upgrade",
                "pip==22.2.2",
                "setuptools==65.4.1",
                "wheel==0.38.4",
            ]
            subprocess.check_call(
                pip_cmd,
                executable=pip_cmd[0],
                stdin=subprocess.DEVNULL,
            )
            self._patch_pip(path)
            logger.info("Done preparing working environment.")
        else:
            logger.debug("Using existing working environment at %s", path)

        lock = FileLock(os.path.join(path, "pipkin.lock"))
        try:
            lock.acquire(timeout=0.05)
        except filelock.Timeout:
            raise UserError(
                "Could not get exclusive access to the working environment. "
                "Is there another pipkin instance running?"
            )
        logger.debug("Received lock on the working environment")

        return lock, path

    def _get_venv_site_packages_path(self) -> str:
        return get_venv_site_packages_path(self._venv_dir)

    def _patch_pip(self, venv_path: str) -> None:
        sp_cmd = [
            get_venv_executable(venv_path),
            "-c",
            "import sysconfig; print(sysconfig.get_paths()['purelib'])",
        ]
        site_packages_path = subprocess.check_output(
            sp_cmd,
            executable=sp_cmd[0],
            stdin=subprocess.DEVNULL,
            universal_newlines=True,
        ).strip()

        pip_init_path = os.path.join(site_packages_path, "pip", "__init__.py")

        patch = """
import os
import pip._vendor.packaging.markers
import pip._vendor.distlib.markers

ENV_VAR_PREFIX = "pip_marker_"

def patch_context(context):
    for name in os.environ:
        if name.startswith(ENV_VAR_PREFIX):
            marker_name = name[len(ENV_VAR_PREFIX):]
            value =  os.environ[name]
            context[marker_name] = value


def patch_context_function(fun):
    def patched_context_function():
        context = fun()
        patch_context(context)
        return context
    
    return patched_context_function

pip._vendor.packaging.markers.default_environment = \
    patch_context_function(pip._vendor.packaging.markers.default_environment)
pip._vendor.distlib.markers.DEFAULT_CONTEXT = \
    patch_context(pip._vendor.distlib.markers.DEFAULT_CONTEXT)        

"""
        logger.info("Patching %r", pip_init_path)
        with open(pip_init_path, "a", encoding="utf-8") as fp:
            fp.write(patch)

    def _clear_venv(self) -> None:
        sp_path = self._get_venv_site_packages_path()
        logger.debug("Clearing %s", sp_path)
        for name in os.listdir(sp_path):
            full_path = os.path.join(sp_path, name)
            if self._is_initial_venv_item(name):
                logger.debug("skipping %r", name)
                continue
            elif os.path.isfile(full_path):
                logger.debug("removing file %r", name)
                os.remove(full_path)
            else:
                logger.debug("removing directory %r", name)
                assert os.path.isdir(full_path)
                shutil.rmtree(full_path)

    def _populate_venv(self, paths: Optional[List[str]] = None, user: bool = False) -> None:
        """paths and user should be used only with list and freeze commands"""
        logger.debug("Start populating venv")
        self._ensure_venv()
        # TODO: try to re-use the state from the previous command executed in the same session.
        assert not (paths and user)
        if user:
            effective_paths = [self._adapter.get_user_packages_path()]
        else:
            effective_paths = paths
        self._clear_venv()
        dist_infos = self._adapter.list_dists(effective_paths)
        for name in dist_infos:
            meta_dir_name, original_path = dist_infos[name]
            self._prepare_dummy_dist(meta_dir_name, original_path)
        logger.debug("Done populating venv")

    def _prepare_dummy_dist(self, meta_dir_name: str, original_path: str) -> None:
        sp_path = self._get_venv_site_packages_path()
        meta_path = os.path.join(sp_path, meta_dir_name)
        os.mkdir(meta_path, 0o755)

        for name in ["METADATA"]:
            content = self._read_dist_meta_file(meta_dir_name, name, original_path)
            with open(os.path.join(meta_path, name), "bw") as meta_fp:
                meta_fp.write(content)

        # INSTALLER is mandatory according to https://www.python.org/dev/peps/pep-0376/
        with open(os.path.join(meta_path, "INSTALLER"), "w", encoding="utf-8") as installer_fp:
            installer_fp.write("pip\n")

        # create dummy RECORD
        with open(os.path.join(meta_path, "RECORD"), "w", encoding=META_ENCODING) as record_fp:
            for name in ["METADATA", "INSTALLER", "RECORD"]:
                record_fp.write(f"{meta_dir_name}/{name},,\n")

    def _read_dist_meta_file(
        self, meta_dir_name: str, file_name: str, original_container_path: str
    ) -> bytes:
        # TODO: add cache
        path = self._adapter.join_path(original_container_path, meta_dir_name, file_name)
        return self._adapter.read_file(path)

    def _compute_venv_path(self) -> str:
        try:
            # try to share the pip-execution-venv among all pipkin-running-venvs created from
            # same base executable
            exe = get_base_executable()
        except Exception:
            exe = sys.executable

        venv_name = hashlib.md5(str((exe, sys.version_info[0:2])).encode("utf-8")).hexdigest()
        return os.path.join(self._get_workspaces_dir(), venv_name)

    def _get_workspaces_dir(self) -> str:
        return os.path.join(self._get_pipkin_cache_dir(), "workspaces")

    def _get_pipkin_cache_dir(self) -> str:
        result = os.path.join(get_user_cache_dir(), "pipkin")
        if sys.platform == "win32":
            # Windows doesn't have separate user cache dir
            result = os.path.join(result, "cache")
        return result

    def _is_initial_venv_item(self, name: str) -> bool:
        return (
            name in INITIAL_VENV_FILES
            or name in INITIAL_VENV_DISTS
            or name.endswith(".dist-info")
            and name.split("-")[0] in INITIAL_VENV_DISTS
        )

    def _get_venv_state(self, root: str = None) -> Dict[str, float]:
        """Returns mapping from meta_dir names to modification timestamps of METADATA files"""
        if root is None:
            root = self._get_venv_site_packages_path()

        result = {}
        for item_name in os.listdir(root):
            if self._is_initial_venv_item(item_name):
                continue

            if item_name.endswith(".dist-info"):
                metadata_full_path = os.path.join(root, item_name, "METADATA")
                assert os.path.exists(metadata_full_path)
                result[item_name] = os.stat(metadata_full_path).st_mtime

        return result

    def _invoke_pip_with_index_args(
        self,
        pip_args: List[str],
        no_mp_org: bool,
        index_url: str,
        extra_index_urls: List[str],
        no_index: bool,
        find_links: Optional[str],
    ):
        if no_index:
            assert find_links
            self._invoke_pip(pip_args + ["--no-index", "--find-links", find_links])
        else:
            proxy = start_proxy(no_mp_org, index_url, extra_index_urls)
            logger.info("Using PipkinProxy at %s", proxy.get_index_url())

            index_args = ["--index-url", proxy.get_index_url()]
            if find_links:
                index_args += ["--find-links", find_links]

            try:
                self._invoke_pip(pip_args + index_args)
            finally:
                proxy.shutdown()

    def _invoke_pip(self, args: List[str]) -> None:
        pip_cmd = [get_venv_executable(self._venv_dir), "-I", "-m", "pip"]

        if not self._tty:
            pip_cmd += ["--no-color"]

        pip_cmd += [
            "--disable-pip-version-check",
            "--trusted-host",
            "127.0.0.1",
        ] + args
        logger.debug("Calling pip: %s", " ".join(shlex.quote(arg) for arg in pip_cmd))

        env = {key: os.environ[key] for key in os.environ if not key.startswith("PIP_")}
        env["PIP_CACHE_DIR"] = self._get_pipkin_cache_dir()

        subprocess.check_call(pip_cmd, executable=pip_cmd[0], env=env, stdin=subprocess.DEVNULL)

    def _compile_with_mpy_cross(
        self, source_path: str, target_path: str, mpy_cross_path: Optional[str]
    ) -> None:
        if mpy_cross_path is None:
            mpy_cross_path = self._ensure_mpy_cross()

        # user-provided executable is assumed to have been validated with proper error messages in main()
        assert os.path.exists
        assert os.access(mpy_cross_path, os.X_OK)
        args = (
            [mpy_cross_path] + self._adapter.get_mpy_cross_args() + ["-o", target_path, source_path]
        )
        subprocess.check_call(args, executable=args[0], stdin=subprocess.DEVNULL)

    def _ensure_mpy_cross(self) -> str:
        impl_name, ver_prefix = self._adapter.get_implementation_name_and_version_prefix()
        path = self._get_mpy_cross_path(impl_name, ver_prefix)
        if not os.path.exists(path):
            self._download_mpy_cross(impl_name, ver_prefix, path)
        return path

    def _download_mpy_cross(
        self, implementation_name: str, version_prefix: str, target_path: str
    ) -> None:
        os.makedirs(os.path.dirname(target_path), exist_ok=True)
        meta_url = f"https://raw.githubusercontent.com/aivarannamaa/pipkin/master/data/{implementation_name}-mpy-cross.json"
        with urlopen(url=meta_url) as fp:
            meta = json.load(fp)

        if version_prefix not in meta:
            raise UserError(f"Can't find mpy-cross for {implementation_name} {version_prefix}")

        version_data = meta[version_prefix]

        if sys.platform == "win32":
            os_marker = "windows"
        elif sys.platform == "darwin":
            os_marker = "macos"
        elif sys.platform == "linux":
            os_marker = "linux"
        else:
            raise AssertionError(f"Unexpected sys.platform {sys.platform}")

        full_marker = f"{os_marker}-{platform.machine()}"

        if full_marker not in version_data:
            raise UserError(
                f"Can't find {full_marker} mpy-cross for {implementation_name} {version_prefix}"
            )

        download_url = version_data[full_marker]

        urllib.request.urlretrieve(download_url, target_path)
        os.chmod(target_path, os.stat(target_path).st_mode | stat.S_IEXEC)

    def _get_mpy_cross_path(self, implementation_name: str, version_prefix: str) -> str:
        basename = f"mpy-cross_{implementation_name}_{version_prefix}"
        if sys.platform == "win32":
            basename += ".exe"

        return os.path.join(self._get_pipkin_cache_dir(), "mpy-cross", basename)

    def _report_progress(self, msg: str, end="\n") -> None:
        if not self._quiet:
            print(msg, end=end)
            sys.stdout.flush()



import logging
import threading
import tkinter as tk
from tkinter import ttk
from turcar import get_workbench
from turcar.languages import tr
# from cefpython3 import cefpython as cef
import sys

logger = logging.getLogger(__name__)


# 浏览器内容窗口
class CefBrowser(ttk.Frame):
    def __init__(self, parent=None):
        self.browser = None
        super().__init__(parent)


class InnerBrowser(ttk.Frame):
    def __init__(self, master):
        ttk.Frame.__init__(self, master, style="ViewBody.TFrame")

        self.text_command = tk.Text(self, width=160, height=1)
        default_text = "http://180.76.235.69"
        self.text_command.insert("1.0", default_text)

        rect = [0, 50, 1200, 1000]
        self.thread = threading.Thread(target=self.embed_browser_thread, args=(self, rect, default_text))
        self.thread.start()
        self.browser_widget = CefBrowser()

    def initialize_cef(self):
        rect = [0, 50, 1200, 1000]
        default_text = "http://180.76.235.69"
        self.thread = threading.Thread(target=self.embed_browser_thread, args=(self, rect, default_text))
        self.thread.start()
        self.browser_widget = CefBrowser()

    def close_cef(self):
        cef.Shutdown()

    def open_programming_exercises(self):
        rect = [0, 0, 800, 600]
        default_text = self.text_command.get("1.0", "end-1c")
        self.embed_browser_thread(self, rect, default_text)

    def embed_browser_thread(self, frame, _rect, url):
        sys.excepthook = cef.ExceptHook
        window_info = cef.WindowInfo(frame.winfo_id())
        window_info.SetAsChild(frame.winfo_id(), _rect)
        cef.Initialize()
        self.browser_widget.browser = cef.CreateBrowserSync(window_info, url=url)
        cef.MessageLoop()
        cef.Shutdown()


def init() -> None:
    get_workbench().add_view(InnerBrowser, tr("programming_exercises"), "se")



"""LED matrix driver"""

from __future__ import annotations

from typing import List

import digitalio
from circuitpython_typing import ReadableBuffer

class PewPew:
    """This is an internal module to be used by the ``pew.py`` library from
    https://github.com/pewpew-game/pew-pewpew-standalone-10.x to handle the
    LED matrix display and buttons on the ``pewpew10`` board.

    Usage::

        This singleton class is instantiated by the ``pew`` library, and
        used internally by it. All user-visible interactions are done through
        that library."""

    def __init__(
        self,
        buffer: ReadableBuffer,
        rows: List[digitalio.DigitalInOut],
        cols: List[digitalio.DigitalInOut],
        buttons: digitalio.DigitalInOut,
    ) -> None:
        """Initializes matrix scanning routines.

        The ``buffer`` is a 64 byte long ``bytearray`` that stores what should
        be displayed on the matrix. ``rows`` and ``cols`` are both lists of
        eight ``DigitalInputOutput`` objects that are connected to the matrix
        rows and columns. ``buttons`` is a ``DigitalInputOutput`` object that
        is connected to the common side of all buttons (the other sides of the
        buttons are connected to rows of the matrix)."""
        ...



import ast

from turcar.plugins.cpython_backend import get_backend


def augment_source(source, cmd):
    if "Flask" not in source:
        # don't bother analyzing further
        return source

    try:
        tree = ast.parse(source)
    except SyntaxError:
        return source

    var_name = None
    found_run = False

    for node in ast.walk(tree):
        if (
            isinstance(node, ast.Assign)
            and isinstance(node.value, ast.Call)
            and isinstance(node.value.func, ast.Name)  # TODO: could be also flask.Flask
            and node.value.func.id == "Flask"
            and len(node.targets) == 1
            and isinstance(node.targets[0], ast.Name)
        ):
            var_name = node.targets[0].id
        elif (
            isinstance(node, ast.Call)
            and isinstance(node.func, ast.Attribute)
            and isinstance(node.func.value, ast.Name)
            and node.func.value.id == var_name
            and node.func.attr == "run"
        ):
            found_run = True

    if not var_name or found_run:
        return source
    else:
        return (
            source
            + """

if "{app_name}" in globals():
    # Conservative options for minimum technical risks.
    # Users who need more control should call run explicitly.
    print(" # Running the app with options chosen by turcar. See Help for details.")
    {app_name}.run(threaded=False, debug=False, use_reloader=False)
""".format(
                app_name=var_name
            )
        )


def load_plugin():
    get_backend().add_source_preprocessor(augment_source)



"""Access GIF-format images
"""

from __future__ import annotations

import typing
from typing import Union

import displayio
from circuitpython_typing import ReadableBuffer

class GifWriter:
    def __init__(
        self,
        file: Union[typing.BinaryIO, str],
        width: int,
        height: int,
        colorspace: displayio.Colorspace,
        loop: bool = True,
        dither: bool = False,
    ) -> None:
        """Construct a GifWriter object

        :param file: Either a file open in bytes mode, or the name of a file to open in bytes mode.
        :param width: The width of the image.  All frames must have the same width.
        :param height: The height of the image.  All frames must have the same height.
        :param colorspace: The colorspace of the image.  All frames must have the same colorspace.  The supported colorspaces are ``RGB565``, ``BGR565``, ``RGB565_SWAPPED``, ``BGR565_SWAPPED``, and ``L8`` (greyscale)
        :param loop: If True, the GIF is marked for looping playback
        :param dither: If True, and the image is in color, a simple ordered dither is applied.
        """
        ...
    def __enter__(self) -> GifWriter:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the hardware when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    def deinit(self) -> None:
        """Close the underlying file."""
        ...
    def add_frame(self, bitmap: ReadableBuffer, delay: float = 0.1) -> None:
        """Add a frame to the GIF.

        :param bitmap: The frame data
        :param delay: The frame delay in seconds.  The GIF format rounds this to the nearest 1/100 second, and the largest permitted value is 655 seconds.
        """
        ...

class OnDiskGif:
    """Loads one frame of a GIF into memory at a time.

    The code can be used in cooperation with displayio but this mode is relatively slow:

    .. code-block:: Python

      import board
      import gifio
      import displayio
      import time

      display = board.DISPLAY
      splash = displayio.Group()
      display.root_group = splash

      odg = gifio.OnDiskGif('/sample.gif')

      start = time.monotonic()
      next_delay = odg.next_frame() # Load the first frame
      end = time.monotonic()
      overhead = end - start

      face = displayio.TileGrid(
          odg.bitmap,
          pixel_shader=displayio.ColorConverter(
              input_colorspace=displayio.Colorspace.RGB565_SWAPPED
          ),
      )
      splash.append(face)
      board.DISPLAY.refresh()

      # Display repeatedly.
      while True:
          # Sleep for the frame delay specified by the GIF,
          # minus the overhead measured to advance between frames.
          time.sleep(max(0, next_delay - overhead))
          next_delay = odg.next_frame()

    The displayio Group and TileGrid layers can be bypassed and the image can
    be directly blitted to the full screen. This can give a speed-up of ~4x to
    ~6x depending on the GIF and display. This requires an LCD that uses
    standard codes to set the update area, and which accepts RGB565_SWAPPED
    pixel data directly:

    .. code-block:: Python

      # Initial set-up the same as above

      # Take over display to drive directly
      display.auto_refresh = False
      display_bus = display.bus

      # Display repeatedly & directly.
      while True:
          # Sleep for the frame delay specified by the GIF,
          # minus the overhead measured to advance between frames.
          time.sleep(max(0, next_delay - overhead))
          next_delay = odg.next_frame()

          display_bus.send(42, struct.pack(">hh", 0, odg.bitmap.width - 1))
          display_bus.send(43, struct.pack(">hh", 0, odg.bitmap.height - 1))
          display_bus.send(44, odg.bitmap)

      # The following optional code will free the OnDiskGif and allocated resources
      # after use. This may be required before loading a new GIF in situations
      # where RAM is limited and the first GIF took most of the RAM.
      odg.deinit()
      odg = None
      gc.collect()

    """

    def __init__(self, file: str) -> None:
        """Create an `OnDiskGif` object with the given file.
        The GIF frames are decoded into RGB565 big-endian format.
        `displayio` expects little-endian, so the example above uses `Colorspace.RGB565_SWAPPED`.

        :param file file: The name of the GIF file.
        """
        ...
    def __enter__(self) -> OnDiskGif:
        """No-op used by Context Managers."""
        ...
    def __exit__(self) -> None:
        """Automatically deinitializes the GIF when exiting a context. See
        :ref:`lifetime-and-contextmanagers` for more info."""
        ...
    width: int
    """Width of the gif. (read only)"""
    height: int
    """Height of the gif. (read only)"""
    bitmap: displayio.Bitmap
    """The bitmap used to hold the current frame."""
    def next_frame(self) -> float:
        """Loads the next frame. Returns expected delay before the next frame in seconds."""
    duration: float
    """Returns the total duration of the GIF in seconds. (read only)"""
    frame_count: int
    """Returns the number of frames in the GIF. (read only)"""
    min_delay: float
    """The minimum delay found between frames. (read only)"""
    max_delay: float
    """The maximum delay found between frames. (read only)"""

    def deinit(self) -> None:
        """Release resources allocated by OnDiskGif."""
        ...



# -*- coding: utf-8 -*-

import tkinter as tk
from tkinter import ttk

from _tkinter import TclError

from turcar import get_runner, get_workbench
from turcar.common import InlineCommand
from turcar.languages import tr
from turcar.memory import MAX_REPR_LENGTH_IN_GRID, MemoryFrame, format_object_id, parse_object_id
from turcar.misc_utils import shorten_repr


class HeapView(MemoryFrame):
    def __init__(self, master):
        MemoryFrame.__init__(self, master, ("id", "value"))

        self.tree.column("id", width=100, anchor=tk.W, stretch=False)
        self.tree.column("value", width=150, anchor=tk.W, stretch=True)

        self.tree.heading("id", text=tr("ID"), anchor=tk.W)
        self.tree.heading("value", text=tr("Value"), anchor=tk.W)

        get_workbench().bind("get_heap_response", self._handle_heap_event, True)

        get_workbench().bind("DebuggerResponse", self._request_heap_data, True)
        get_workbench().bind("ToplevelResponse", self._request_heap_data, True)
        # Showing new globals may introduce new interesting objects
        get_workbench().bind("get_globals_response", self._request_heap_data, True)

        self.bind("<Map>", self._on_map, True)
        self.bind("<Unmap>", self._on_unmap, True)
        info_icon = get_workbench().get_image("information")
        self.info_label = ttk.Label(
            get_workbench().get_toolbar(),
            text=tr("Heap mode is on.\nClose Heap view to turn it off."),
            image=info_icon,
            compound="left",
            background="#F4FFBA",
            foreground="green",
            padding=(3, 0),
        )

    def _update_data(self, data):
        self._clear_tree()
        for value_id in sorted(data.keys()):
            node_id = self.tree.insert("", "end")
            self.tree.set(node_id, "id", format_object_id(value_id))
            self.tree.set(
                node_id, "value", shorten_repr(data[value_id].repr, MAX_REPR_LENGTH_IN_GRID)
            )

    def before_show(self):
        self._request_heap_data(even_when_hidden=True)

    def get_object_id(self):
        iid = self.tree.focus()
        if iid != "":
            return parse_object_id(self.tree.item(iid)["values"][0])

        return None

    def _request_heap_data(self, msg=None, even_when_hidden=False):
        if self.winfo_ismapped() or even_when_hidden:
            # TODO: update itself also when it becomes visible
            if get_runner() is not None:
                get_runner().send_command(InlineCommand("get_heap"))

    def _handle_heap_event(self, msg):
        if self.winfo_ismapped():
            if hasattr(msg, "heap"):
                self._update_data(msg.heap)

    def _on_map(self, event):
        self.info_label.grid(row=0, column=1005)

    def _on_unmap(self, event):
        try:
            self.info_label.grid_remove()
        except TclError:
            pass


def load_plugin() -> None:
    get_workbench().add_view(HeapView, tr("Heap"), "e")



"""
Module: '_api.distancesensor' on micropython-v1.14-893-lego learning system hub
"""
# MCU: {'machine': 'LEGO Technic Large Hub with STM32F413xx', 'sysname': 'LEGO Technic Large Hub', 'platform': 'LEGO Learning System Hub', 'nodename': 'LEGO Learning System Hub', 'ver': 'v1.14-893', 'release': '1.14.0', 'name': 'micropython', 'family': 'micropython', 'port': 'LEGO Learning System Hub', 'version': '1.14.0', 'mpy': 517, 'build': '893'}
# Stubber: 1.7.2
from typing import Any


class DistanceSensor():
    def __init__(self, *argv, **kwargs) -> None:
        """"""

    CM = 'cm' # type: str
    IN = 'in' # type: str
    PERCENT = '%' # type: str
    def get_distance_cm(self, *args, **kwargs) -> Any:
        """"""

    def get_distance_inches(self, *args, **kwargs) -> Any:
        """"""

    def get_distance_percentage(self, *args, **kwargs) -> Any:
        """"""

    def light_up(self, *args, **kwargs) -> Any:
        """"""

    def light_up_all(self, *args, **kwargs) -> Any:
        """"""

    def wait_for_distance_closer_than(self, *args, **kwargs) -> Any:
        """"""

    def wait_for_distance_farther_than(self, *args, **kwargs) -> Any:
        """"""

LPF2_FLIPPER_DISTANCE = 62 # type: int
PORTS = {} # type: dict
def clamp(*args, **kwargs) -> Any:
    """"""

def is_type(*args, **kwargs) -> Any:
    """"""

def newSensorDisconnectedError(*args, **kwargs) -> Any:
    """"""

def sleep_ms(*args, **kwargs) -> Any:
    """"""




class Battery:
    """
    A stub class to represent the battery member of the EV3Brick class.
    """

    def __init__(self):
        ...

    def voltage(self) -> int:
        """
        Gets the voltage of the battery.

        Returns:
        Voltage in mV.
        """
        return 0

    def current(self) -> int:
        """
        Gets the current supplied by the battery.

        Returns:
        Battery current in mA.
        """
        return 0



from pybricks.parameters import Port


class Ev3devSensor:
    """
    Read values of an ev3dev-compatible sensor.

    Args:
        port (Port): Port to which the device is connected.

    Attributes:
        sensor_index (int): Index of the ev3dev sysfs lego-sensor class.
        port_index (int): Index of the ev3dev sysfs lego-port class.
    """

    def __init__(self, port: Port):
        self.sensor_index = 0  # type: int
        self.port_index = 0  # type: int
        if port == Port.A or port == Port.B or port == Port.C or port == Port.D:
            raise ValueError("Sensors must use Port S1, S2, S3, or S4.")

    def read(self, mode: str) -> tuple:
        """
        Reads values at a given mode.

        Args:
            mode (str): Mode name.

        Returns:
            Values read from the sensor.
        """
        return ()



# -*- coding: utf-8 -*-
from turcar import get_workbench
from turcar.languages import tr
from turcar.ui_utils import select_sequence


def load_plugin() -> None:
    def create_edit_command_handler(virtual_event_sequence):
        def handler(event=None):
            widget = get_workbench().focus_get()
            if widget:
                return widget.event_generate(virtual_event_sequence)

            return None

        return handler

    get_workbench().add_command(
        "undo",
        "edit",
        tr("Undo"),
        create_edit_command_handler("<<Undo>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-z>", "<Command-z>"),
        extra_sequences=["<Control-Greek_zeta>"],
        skip_sequence_binding=True,
        group=10,
    )

    get_workbench().add_command(
        "redo",
        "edit",
        tr("Redo"),
        create_edit_command_handler("<<Redo>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-y>", "<Command-y>"),
        extra_sequences=[
            select_sequence("<Control-Shift-Z>", "<Command-Shift-Z>"),
            "<Control-Greek_upsilon>",
            "<Control-Shift-Greek_ZETA>",
        ],
        skip_sequence_binding=True,
        group=10,
    )

    get_workbench().add_command(
        "Cut",
        "edit",
        tr("Cut"),
        create_edit_command_handler("<<Cut>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-x>", "<Command-x>"),
        extra_sequences=["<Control-Greek_chi>"],
        skip_sequence_binding=True,
        group=20,
    )

    get_workbench().add_command(
        "Copy",
        "edit",
        tr("Copy"),
        create_edit_command_handler("<<Copy>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-c>", "<Command-c>"),
        extra_sequences=["<Control-Greek_psi>"],
        skip_sequence_binding=True,
        group=20,
    )

    get_workbench().add_command(
        "Paste",
        "edit",
        tr("Paste"),
        create_edit_command_handler("<<Paste>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-v>", "<Command-v>"),
        extra_sequences=["<Control-Greek_omega>"],
        skip_sequence_binding=True,
        group=20,
    )

    get_workbench().add_command(
        "SelectAll",
        "edit",
        tr("Select all"),
        create_edit_command_handler("<<SelectAll>>"),
        tester=None,  # TODO:
        default_sequence=select_sequence("<Control-a>", "<Command-a>"),
        extra_sequences=["<Control-Greek_alpha>"],
        skip_sequence_binding=True,
        group=20,
    )
